"""A request model emulates customers interest on booking flight tickets.
defined by a set of parameters that define the rate of requests per day 
as the departure date approaches and the amount of seats for each request.
"""

from dataclasses import dataclass
import json
from typing import List, Union
from rmlab._data.conversions import parse_data_from_json
from rmlab._data.enums import CustomersModelKind, FileExtensions
from rmlab.data.parametric._customers_base import CustomersModel


@dataclass
class BaseRequestModel(CustomersModel):
    """Base dataclass for request models."""

    pass


@dataclass
class ExponentialPoisson(BaseRequestModel):
    """A particular kind of request model, characterized by an exponential curve that
    for a given day before departure (*d* onwards, from 0 *departure day* to +D *on-sale day*) returns the cumulative booking requests for that day

    $$
    CumulativeRequestsOn(d) = scale · \exp(-d/ \\beta)
    $$

    On a given day before departure *d*, the average booking requests is given by

    $$
    \lambda_{requests}(d) = CumulativeRequestsOn(d) - CumulativeRequestsOn(d+1).
    $$

    Note that $d+1$ is the day *before* $d$.

    From $\lambda_{requests}$ we sample the number of booking requests on day before departure $d$ as:

    $$
    \# booking\,requests(d) \sim Poisson(\lambda_{requests}(d))
    $$

    For each booking request, we also sample the requested seats from a zero-truncated Poisson

    $$
    \# seats\,per\,request \sim Poisson_{ZT}(\lambda_{seats})
    $$

    Args:
      class_name (str): Customers class
      beta (float): $$\\beta$$
      scale (float): $$scale$$
      lambda_seats (float): $$\lambda_{seats}$$

    Examples:
    ```py
    bus_req_model = ExponentialPoisson(class_name="business", beta=30, scale=300, lambda_seats=1.5)
    ```

    JSON representation:
    ```json
    {
      "business": {"type": "poisson", "beta": 30.0, "scale": 400, "lambda_seats": 1.5},
    }
    ```
    """

    class_name: str
    beta: float
    scale: float
    lambda_seats: float


@dataclass
class MultipleRequestModels(BaseRequestModel):
    """An aggregation of request models for different customer classes.

    Example:
    ```py
    bus_req_model = ExponentialPoisson(class_name="business", beta=30, scale=300, lambda_seats=1.5)
    lei_req_model = ExponentialPoisson(class_name="leisure", beta=10, scale=400, lambda_seats=3)

    mul_req_model = MultipleRequestModels(bus_req_model, lei_req_model)
    ```

    JSON representation:
    ```json
    {
      "business": {"type": "poisson", "beta": 30.0, "scale": 400, "lambda_seats": 1.5},
      "leisure": {"type": "poisson", "beta": 10.0, "scale": 300, "lambda_seats": 3}
    }
    ```
    """

    models: List[BaseRequestModel]


def make_customers_request_model_from_json(
    filename_or_dict: Union[str, dict]
) -> BaseRequestModel:
    """Make a request model instance from a json representation (from file or dict).


    Args:
        filename_or_dict (Union[str, dict]): JSON filename or dictionary in json format

    Raises:
        ValueError: If JSON is not a dict
        ValueError: If model type is invalid
        ValueError: If JSON content is empty

    Returns:
        A customers request model instance

    Example from dict:
    ```py
    my_dict = dict()
    my_dict["business"] = {"type": "poisson", "beta": 30.0, "scale": 400, "lambda_seats": 1.5}
    my_dict["leisure"] = {"type": "poisson", "beta": 10.0, "scale": 300, "lambda_seats": 3}

    my_req_model = make_from_json(my_dict)
    ```

    Example from file:
    `my_req_model.json`
    ```json
    {
      "business": {"type": "poisson", "beta": 30.0, "scale": 400, "lambda_seats": 1.5},
      "leisure": {"type": "poisson", "beta": 10.0, "scale": 300, "lambda_seats": 3}
    }
    ```
    ```py
    my_req_model = make_customers_request_model_from_json("my_req_model.json")
    ```
    """

    model_id, filename, md5hash, content = parse_data_from_json(filename_or_dict)

    if not isinstance(content, dict):
        raise ValueError(f"Expected dict format in {filename_or_dict}")

    content: dict

    if len(content) > 0:
        models = list()
        for class_name, fields in content.items():
            if fields["type"] == "poisson":
                models.append(
                    ExponentialPoisson(
                        id=model_id,
                        filename=filename,
                        cnt=json.dumps(content),
                        hash=md5hash,
                        kind=CustomersModelKind.REQUEST,
                        extension=FileExtensions.JSON,
                        class_name=class_name,
                        beta=fields["beta"],
                        scale=fields["scale"],
                        lambda_seats=fields["lambda_seats"],
                    )
                )
            else:
                raise ValueError(f'Unknown request model type {fields["type"]}')

        if len(models) > 1:
            return MultipleRequestModels(
                id=model_id,
                filename=filename,
                cnt=json.dumps(content),
                hash=md5hash,
                kind=CustomersModelKind.REQUEST,
                extension=FileExtensions.JSON,
                models=models,
            )
        else:
            return models[0]

    else:
        raise ValueError(f"Empty content")



"""Interface for uploading pricing models data."""

import os
from typing import List, Optional
from rmlab._api.upload import APIUploadInternal
from rmlab._data.enums import (
    CustomersModelKind,
    ParametricModelKind,
)


class APIUploadCustomersModels(APIUploadInternal):
    """Exposes functions for uploading customers models data to the server."""

    async def upload_customers_request_model(
        self,
        scen_id: int,
        data_fn: str,
    ) -> None:
        """Upload to server a parametric model defined in file, modelling how customers request books.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename (.json) defining the parametric model
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.CUSTOMERS,
            kind=CustomersModelKind.REQUEST,
            data_fn=data_fn,
        )

    async def upload_customers_choice_model(
        self,
        scen_id: int,
        data_fn: str,
    ) -> None:
        """Upload to server a parametric model defined in file, modelling how customers choose between competing book offers.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename (.json) defining the parametric model
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.CUSTOMERS,
            kind=CustomersModelKind.CHOICE,
            data_fn=data_fn,
        )

    async def upload_batch_customers_models(
        self,
        scen_id: int,
        *,
        request_models_fns: Optional[List[str]] = None,
        choice_models_fns: Optional[List[str]] = None,
    ):
        if request_models_fns is None:
            request_models_fns = list()
        if choice_models_fns is None:
            request_models_fns = list()

        not_existing_fns = [
            fn for fn in request_models_fns if fn and not os.path.exists(fn)
        ] + [fn for fn in choice_models_fns if fn and not os.path.exists(fn)]

        if len(not_existing_fns) > 0:
            raise FileNotFoundError(not_existing_fns)

        for crm in request_models_fns:

            await self._upload_parametric_model(
                scen_id, ParametricModelKind.CUSTOMERS, CustomersModelKind.REQUEST, crm
            )

        for ccm in choice_models_fns:

            await self._upload_parametric_model(
                scen_id, ParametricModelKind.CUSTOMERS, CustomersModelKind.CHOICE, ccm
            )



import os, logging
from typing import Optional, Union

from rmlab_http_client.types import FileExtensionType

from rmlab._api.base import APIBaseInternal
from rmlab._data.conversions import (
    C2S_FlightDataName2ConvertFunction,
    UploadableDataClassName2DataKind,
    UploadableFlightDataType,
)
from rmlab._data.enums import (
    CustomersModelKind,
    ParametricModelKind,
    PricingModelKind,
)
from rmlab._data.types import (
    BoundedItem,
    UnboundedItem,
)
from rmlab.data.parametric.customers_choice import make_customers_choice_model_from_json
from rmlab.data.parametric.customers_request import (
    make_customers_request_model_from_json,
)
from rmlab.data.parametric.pricing_behavior import make_pricing_behavior_from_json
from rmlab.data.parametric.pricing_optimizer import make_pricing_optimizer_from_json
from rmlab.data.parametric.pricing_range import make_pricing_range_from_json


_Logger = logging.getLogger(__name__)


class APIUploadInternal(APIBaseInternal):
    """Interface to upload data to server"""

    async def _upload_bounded_items(
        self, scen_id: int, category: BoundedItem, data_fn: str
    ) -> None:
        """Upload a set of bounded items defined in a file.

        Args:
            scen_id (int): Scenario ID
            category (BoundedItem): Category
            data_fn (str): Filename (.csv or .json) defining the items

        Raises:
            ValueError: If category is not bounded
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
        """

        if not issubclass(category, BoundedItem):
            raise ValueError(
                f"Category `{category.__name__}` must inherit from `BoundedItem`"
            )

        if not os.path.exists(data_fn):
            raise FileNotFoundError(data_fn)

        _, ext = os.path.splitext(data_fn)
        ext: str = FileExtensionType.str_to_enum_value(ext.replace(".", "")).value

        _Logger.debug(f"Uploading `{category.__name__}` items")

        await self._submit_call(
            "api-data-bounded-post",
            scen_id=scen_id,
            category=category.__name__.lower(),
            file=data_fn,
            extension=ext,
        )

        _Logger.debug(f"File `{data_fn}` of category `{category.__name__}` uploaded")

    async def _upload_unbounded_items(
        self,
        scen_id: int,
        category: UnboundedItem,
        data_fn: str,
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> None:
        """Upload a set of unbounded items defined in a file.
        Used to upload a set of flight schedules or a set of flights to server.

        Arguments `citysector_id` and `sector_id` are always optional. If any of them provided, concurrent uploads of flights
            files belonging to different citysectors/sectors are allowed.

        Args:
            scen_id (int): Scenario ID
            category (UnboundedItem): Category
            data_fn (str): Filename (.csv or .json) defining the items
            citysector_id (Optional[str], optional): Citysector ID to which items in file are associated. Defaults to None.
            sector_id (Optional[str], optional): Sector ID to which items in file are associated. Defaults to None.

        Raises:
            ValueError: If category is not unbounded
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
        """

        if not issubclass(category, UnboundedItem):
            raise ValueError(
                f"Category `{category.__name__}` must inherit from `UnboundedItem`"
            )

        _, ext = os.path.splitext(data_fn)
        ext: str = FileExtensionType.str_to_enum_value(ext.replace(".", "")).value

        _Logger.debug(f"Uploading `{category.__name__}` items")

        await self._submit_call(
            "api-data-unbounded-post",
            scen_id=scen_id,
            category=category.__name__.lower(),
            citysector_id=citysector_id,
            sector_id=sector_id,
            file=data_fn,
            extension=ext,
        )

        _Logger.debug(f"File `{data_fn}` of category `{category.__name__}` uploaded")

    async def _upload_flight_data(
        self,
        scen_id: int,
        flight_data: UploadableFlightDataType,
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> None:
        """Upload flight data to server. Data types can be `FlightDataBooks`, `FlightDataThresholdSettings` or `FlightDataPricePerSeatSettings`.
        Any previous existing data is overwritten.
        At least `citysector_id` and/or `sector_id` associated to the flight must be defined.

        Args:
            scen_id (int): Scenario ID
            flight_data (UploadableFlightDataType): Flight data to upload.
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Target sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined
            ValueError: If flight data has incorrect type.
        """

        if citysector_id is None and sector_id is None:
            raise ValueError(
                f"Require definition of either `citysector_id` or `sector_id`"
            )

        flight_data_class_name = type(flight_data).__name__

        if flight_data_class_name not in UploadableDataClassName2DataKind:
            raise ValueError(
                f"Flight data of type `{type(flight_data)}` cannot be uploaded"
            )

        flight_id, data = C2S_FlightDataName2ConvertFunction[flight_data_class_name](
            flight_data
        )

        kind = UploadableDataClassName2DataKind[flight_data_class_name]

        await self._submit_call(
            "api-data-flight-post",
            scen_id=scen_id,
            flight_id=flight_id,
            kind=kind,
            citysector_id=citysector_id,
            sector_id=sector_id,
            data=data,
        )

    async def _upload_parametric_model(
        self,
        scen_id: int,
        parametric_kind: ParametricModelKind,
        kind: Union[CustomersModelKind, PricingModelKind],
        data_fn: str,
    ) -> None:
        """Upload a parametric model (a `customers` or a `pricing` model) defined in a file to server.

        Args:
            scen_id (int): Scenario ID
            parametric_kind (ParametricModelKind): Parametric kind value (a `customers` or `pricing` parametric model)
            kind (Union[CustomersModelKind, PricingModelKind]): A subkind of parametric model.
            Given parametric_kind=`CUSTOMERS`, any of `REQUEST` or `CHOICE`
            Given parametric_kind=`PRICING`, any of `RANGE`, `BEHAVIOR`, `OPTIMIZER`
            data_fn (str): Filename (.json) defining the parametric model.
        """

        # Make equivalent parametric models types to ensure correctness before uploading to server
        if parametric_kind == ParametricModelKind.CUSTOMERS:
            if kind == CustomersModelKind.REQUEST:
                make_customers_request_model_from_json(data_fn)
            elif kind == CustomersModelKind.CHOICE:
                make_customers_choice_model_from_json(data_fn)
            else:
                raise ValueError(f"Unexpected value of `{kind}`")
        elif parametric_kind == ParametricModelKind.PRICING:
            if kind == PricingModelKind.RANGE:
                make_pricing_range_from_json(data_fn)
            elif kind == PricingModelKind.BEHAVIOR:
                make_pricing_behavior_from_json(data_fn)
            elif kind == PricingModelKind.OPTIMIZER:
                make_pricing_optimizer_from_json(data_fn)
            else:
                raise ValueError(f"Unexpected value of `{kind}`")
        else:
            raise ValueError(f"Unexpected value of `{parametric_kind}`")

        _, ext = os.path.splitext(data_fn)
        ext: str = FileExtensionType.str_to_enum_value(ext.replace(".", "")).value

        _Logger.debug(f"Uploading `{parametric_kind.value}` `{kind.value}` model")

        pmodel_id = os.path.basename(data_fn)

        await self._submit_call(
            "api-data-pmodel-post",
            scen_id=str(scen_id),
            kind=parametric_kind.value + "_" + kind.value,
            pmodel_id=pmodel_id,
            file=data_fn,
            extension=ext,
        )

        _Logger.debug(f"File `{data_fn}` uploaded as pmodel with id `{pmodel_id}`")



import hashlib
import json
import os
from typing import Any, Callable, List, Mapping, Tuple, Union

from dataclasses import asdict
import uuid

from rmlab._data.enums import FlightDataKind, FlightEvent
from rmlab.data.flight import (
    FlightData,
    FlightDataBooks,
    FlightDataEvents,
    FlightDataForecastedBooks,
    FlightDataPricePerSeatSettings,
    FlightDataThresholdSettings,
)


# TODO DEPRECATED
# DataKind2DataClass: Mapping[FlightDataKind, FlightData] = {
#     FlightDataKind.ACTUAL_BOOKS: FlightDataBooks,
#     FlightDataKind.EXPECTED_BOOKS: FlightDataForecastedBooks,
#     FlightDataKind.DYNAMIC_BOOKS: FlightDataForecastedBooks,
#     FlightDataKind.THRESHOLDS_SETTINGS: FlightDataThresholdSettings,
#     FlightDataKind.PRICE_PER_SEAT_SETTINGS: FlightDataPricePerSeatSettings,
#     FlightDataKind.EVENTS: FlightDataEvents,
# }

UploadableFlightDataType = Union[
    FlightDataBooks, FlightDataThresholdSettings, FlightDataPricePerSeatSettings
]

UploadableDataClassName2DataKind: Mapping[str, FlightDataKind] = {
    FlightDataBooks.__name__: FlightDataKind.ACTUAL_BOOKS,
    FlightDataThresholdSettings.__name__: FlightDataKind.THRESHOLDS_SETTINGS,
    FlightDataPricePerSeatSettings.__name__: FlightDataKind.PRICE_PER_SEAT_SETTINGS,
}

# ---- Server to Client format conversions


def _s2c_conversion_actual_books(flight_id: str, response: dict) -> FlightDataBooks:
    return FlightDataBooks(
        id=flight_id,
        timestamps_array=response["Timestamp"],
        fares_array=response["Fares"],
        pps_array=response["PPS"],
        seats_array=response["Seats"],
        cumulated_seats_array=response["CumSeats"],
        cumulated_revenue_array=response["CumRevenue"],
    )


def _s2c_conversion_future_books(
    flight_id: str, response: dict
) -> FlightDataForecastedBooks:
    return FlightDataForecastedBooks(
        id=flight_id,
        timestamps_array=response["Timestamp"],
        seats_array=response["AvgSeats"],
        pps_array=response["AvgPPS"],
    )


def _s2c_conversion_thresholds_settings(
    flight_id: str, response: dict
) -> FlightDataThresholdSettings:

    args: Mapping[str, List[int]] = dict()

    for fare_key, fare_threshold in response.items():
        if fare_key != "Timestamp":
            args[fare_key] = fare_threshold

    return FlightDataThresholdSettings(
        id=flight_id,
        timestamps_array=response["Timestamp"],
        fare_to_threshold_array=args,
    )


def _s2c_conversion_price_per_seat_settings(
    flight_id: str, response: dict
) -> FlightDataPricePerSeatSettings:

    args: Mapping[str, List[int]] = dict()

    for fare_key, fare_pps in response.items():
        if fare_key != "Timestamp":
            args[fare_key] = fare_pps

    return FlightDataPricePerSeatSettings(
        id=flight_id, timestamps_array=response["Timestamp"], fare_to_pps_array=args
    )


def _s2c_conversion_events(flight_id: str, response: dict) -> FlightDataEvents:

    return FlightDataEvents(
        id=flight_id,
        timestamps_array=response["Timestamp"],
        events_array=[FlightEvent.str_to_enum_value(e) for e in response["Kind"]],
    )


S2C_DataKind2ConvertFunction: Mapping[
    FlightDataKind, Callable[[str, dict], FlightData]
] = {
    FlightDataKind.ACTUAL_BOOKS: _s2c_conversion_actual_books,
    FlightDataKind.EXPECTED_BOOKS: _s2c_conversion_future_books,
    FlightDataKind.DYNAMIC_BOOKS: _s2c_conversion_future_books,
    FlightDataKind.THRESHOLDS_SETTINGS: _s2c_conversion_thresholds_settings,
    FlightDataKind.PRICE_PER_SEAT_SETTINGS: _s2c_conversion_price_per_seat_settings,
    FlightDataKind.EVENTS: _s2c_conversion_events,
}


# ---- Client to Server format conversions


def _c2s_conversion_actual_books(
    data: FlightDataBooks,
) -> Tuple[str, Mapping[str, List[int]]]:

    d = asdict(data)
    id = d.pop("id")
    return id, d


def _c2s_conversion_thresholds_settings(
    data: FlightDataThresholdSettings,
) -> Tuple[str, Mapping[str, List[int]]]:

    d = asdict(data)
    id = d.pop("id")
    return id, d


def _c2s_conversion_price_per_seat_settings(
    data: FlightDataPricePerSeatSettings,
) -> Tuple[str, Mapping[str, List[int]]]:

    d = asdict(data)
    id = d.pop("id")
    return id, d


C2S_FlightDataName2ConvertFunction: Mapping[
    str, Callable[[FlightData], Tuple[str, Mapping[str, List[int]]]]
] = {
    FlightDataBooks.__name__: _c2s_conversion_actual_books,
    FlightDataThresholdSettings.__name__: _c2s_conversion_thresholds_settings,
    FlightDataPricePerSeatSettings.__name__: _c2s_conversion_price_per_seat_settings,
}


def parse_data_from_json(
    filename_or_var: Union[str, dict, list]
) -> Tuple[str, str, str, Union[dict, list]]:
    """Parse data in json format returning a tuple of (identifier, filename, content)

    Args:
        filename_or_var (Union[str, dict, list]): Filename or instance of `format`

    Raises:
        FileNotFoundError: If filename is not found
        ValueError: If type is not recognized

    Returns:
        Tuple[str, str, Mapping[str, Mapping[str, Any]]]: Tuple with
            model ID, source filename, md5hash of content, and content
    """

    if isinstance(filename_or_var, str):
        if os.path.exists(filename_or_var):
            with open(filename_or_var, "r") as f:
                content = json.loads(f.read())
            model_id = (os.path.basename(filename_or_var).split(".")[-2],)
            filename = filename_or_var
        else:
            raise FileNotFoundError(filename_or_var)
    elif isinstance(filename_or_var, list) or isinstance(filename_or_var, dict):
        content = filename_or_var
        model_id = str(uuid.uuid4())[:16]
        filename = "none"
    else:
        raise ValueError(f"Unhandled type {type(filename_or_var)}")

    return (
        model_id,
        filename,
        hashlib.md5(json.dumps(content).encode()).hexdigest(),
        content,
    )



from rmlab.api.remove import APIRemove
from rmlab.api.operations.simulation import APISimulation
from rmlab.api.operations.optimization import APIOptimization
from rmlab.api.upload.core import APIUploadCore
from rmlab.api.upload.parametric.filters import APIUploadParametric
from rmlab.api.upload.parametric.customers import APIUploadCustomersModels
from rmlab.api.upload.parametric.pricing import APIUploadPricingModels
from rmlab.api.upload.flight_data import APIUploadFlightData
from rmlab.api.fetch.core import APIFetchCore
from rmlab.api.fetch.flight_data import APIFetchFlightData
from rmlab.api.observability import APIObservability

class API(
    APISimulation,
    APIOptimization,
    APIUploadCore,
    APIUploadParametric,
    APIUploadCustomersModels,
    APIUploadPricingModels,
    APIUploadFlightData,
    APIRemove,
    APIFetchCore,
    APIFetchFlightData,
    APIObservability,
):
    """Complete mixin interface to run simulations, upload local data and fetch/remove remote data"""

    pass



"""A pricing behavior defines the strategy under which flights assign prices to seats."""

import json
from typing import Union
from dataclasses import dataclass
from rmlab._data.conversions import parse_data_from_json
from rmlab._data.enums import FileExtensions, PricingModelKind
from rmlab.data.parametric._pricing_base import PricingModel


@dataclass
class BaseBehaviorModel(PricingModel):
    """Base class for all types of pricing behavior models."""

    pass


@dataclass
class SeatThresholds(BaseBehaviorModel):
    """A kind of pricing behavior model in which a set of seat thresholds are used to allocate seats to fares based on the flight occupation.

    Args:
      seats_per_dbd (float): Rate of increase in number of seats per day of all thresholds
      zero_shift (float): Per-unit relative displacement of all the thresholds
      squeeze (float): Per-unit relative contraction of the all thresholds

    Example
    ```py
    behavior = SeatThresholds(seats_per_dbd=0.15, zero_shift=0.06, squeeze=0.7)
    ```

    """

    seats_per_dbd: float
    zero_shift: float
    squeeze: float


def make_pricing_behavior_from_json(
    filename_or_dict: Union[str, dict]
) -> BaseBehaviorModel:
    """Make a pricing behavior instance from a json representation (from file or dict).

    Args:
        filename_or_dict (Union[str, dict]): JSON filename or dictionary in json format

    Examples from dict:
    ```py
    dict_behavior = dic()
    dict_behavior["type"] = "rake_straight"
    dict_behavior["seats_per_dbd"] = 0.15
    dict_behavior["zero_shift"] = 0.06
    dict_behavior["squeeze"] = 0.7
    my_pricing_behavior = make_from_json(dict_behavior)
    ```

    Example from file:
    `my_pricing_behavior.json`
    ```json
    {
      "type" : "rake_straight",
      "seats_per_dbd" : 0.15,
      "zero_shift" : 0.06,
      "squeeze" : 0.7
    }
    ```

    ```py
    my_pricing_behavior = make_pricing_behavior_from_json("my_pricing_behavior.json")
    ```

    """

    model_id, filename, md5hash, content = parse_data_from_json(filename_or_dict)

    if not isinstance(content, dict):
        raise TypeError(f"Expected dict format in {filename_or_dict}")

    content: dict

    if content["type"] == "rake_straight":

        return SeatThresholds(
            id=model_id,
            filename=filename,
            cnt=json.dumps(content),
            hash=md5hash,
            kind=PricingModelKind.BEHAVIOR,
            extension=FileExtensions.JSON,
            seats_per_dbd=float(content["seats_per_dbd"]),
            zero_shift=float(content["zero_shift"]),
            squeeze=float(content["squeeze"]),
        )

    else:

        raise ValueError(f'Unknown pricing behavior type `{content["type"]}`')



from dataclasses import dataclass
from rmlab._data.enums import FileExtensions, PricingModelKind
from rmlab.data.items import PModel


@dataclass
class PricingModel(PModel):
    """Pricing models docstring"""

    kind: PricingModelKind
    extension: FileExtensions



"""
Parametric filters assign parametric models to data items.
"""

from hashlib import md5
import json
from typing import List, Union
from dataclasses import dataclass

from rmlab._data.conversions import parse_data_from_json
from rmlab._data.enums import ParametricFilterClassKind
from rmlab.data.items import PModel
from rmlab._data.types import (
    BoundedItem,
    CoreItem,
)


@dataclass
class PFilter(CoreItem, BoundedItem):
    """
    Construct parametric models to items associations.

    Args:
        filter_class (str): Specifies the fields of the target item that are inspected upon filtering matching.
        filter_content (str): The content to be matched
        pmodel_id (str): A pointer (identifier) to the model which stores the set of parameters that we want to assign

    Following examples explain how to associate pmodels to markets and flights.
    In these cases, there is a hierachy that is obeyed when multiple models match multiple
    filtering criteria for a given category item.

    Examples:

    ## Assigning customers models to *citysectors* and *cityroutes*

    Customers models can be assigned either to:

    * **citysectors**: A city pair with defined directionality (eg: a *CitySector* with ID CityA->CityB).

    * **cityroutes**: A city pair with undefined directionality (eg: a *CityRoute* with ID CityA<>CityB).

    Following examples show different filters to associate a given `my_customers_model.json` to:

    **All citysectors:**
    ```py
    pf = ParametricFilter(
        filter_class="network",
        filter_content="",
        pmodel_id="my_customers_model.json")
    ```

    **A cityroute with ID `CityA<>CityB`:**
    ```py
    pf = ParametricFilter(
        filter_class="network",
        filter_content="CityA<>CityB",
        pmodel_id="my_customers_model.json")
    ```

    ## Assigning pricing models models to *flights*

    First, note that:

    * A schedule represents a set of flights departing within a date period,
    referring to a given airline and sector.

    * A sector represents a pair of airports with defined directionality
    (AirportA-> AirportB). We use IATA coding to identify them (eg AAABBB)

    * A route represents a pair of airports with undefined directionality (AirportA <> AirportB).
    We use <>-separated IATA coding to identify them (eg AAA<>BBB)

    Following examples show how to create parametric filters to associate a given `my_pricing_model.json` to:

    **All flights of all schedules:**

    ```py
    pf = ParametricFilter(
        filter_class="network",
        filter_content="",
        pmodel_id="my_pricing_model.json")
    ```

    **All flights of cityroute `CityA<>CityB`:**

    ```py
    pf = ParametricFilter(
        filter_class="cityroute",
        filter_content="CityA<>CityB",
        pmodel_id="my_pricing_model.json")
    ```

    **All flights of citysector `CityA->CityB`:**

    ```py
    pf = ParametricFilter(
        filter_class="cityroute",
        filter_content="CityA->CityB",
        pmodel_id="my_pricing_model.json")
    ```

    **All flights of route `AAA<>BBB`:**

    ```py
    pf = ParametricFilter(
        filter_class="route",
        filter_content="AAA<>BBB",
        pmodel_id="my_pricing_model.json")
    ```

    **All flights of sector `AAABBB`:**

    ```py
    pf = ParametricFilter(
        filter_class="sector",
        filter_content="AAABBB",
        pmodel_id="my_pricing_model.json")
    ```

    **All flights of airline `MyAirline`:**

    ```py
    pf = ParametricFilter(
        filter_class="airline",
        filter_content="MyAirline",
        pmodel_id="my_pricing_model.json")
    ```

    **All flights with flight number `2744`:**

    ```py
    pf = ParametricFilter(
        filter_class="flight_number",
        filter_content="2744",
        pmodel_id="my_pricing_model.json")
    ```

    **All flights whose departure dates are fully contained within a given period `2022/05/24-2022/12/31`:**

    ```py
    pf = ParametricFilter(
        filter_class="in_period",
        filter_content="20220524-20221231",
        pmodel_id="my_pricing_model.json")
    ```

    **All flights sharing several attributes:**

    * whose departure dates are fully contained within a given period `2022/05/24-2022/12/31`
    * belonging to an airline `MyAirline`
    * covering citysector `CityA->CityB`

    ```py
    pf = ParametricFilter(
        filter_class=["in_period", "airline", "citysector"],
        filter_content=["20220524-20221231", "MyAirline", "CityA->CityB"],
        pmodel_id="my_pricing_model.json")
    ```

    All previous examples work as json representations too:
    ```json
    {
      "filter_class": "String",
      "filter_content": "String",
      "pmodel_id": "String"
    }
    ```
    """

    filter_class: List[ParametricFilterClassKind]
    filter_content: List[str]
    pmodel_id: str

    def __post_init__(self):

        if isinstance(self.filter_class, str):
            self.filter_class = [self.filter_class]

        if isinstance(self.filter_content, str):
            self.filter_content = [self.filter_content]

        for i, fc in enumerate(self.filter_class):
            if isinstance(fc, str):
                self.filter_class[i] = ParametricFilterClassKind.str_to_enum_value(fc)

        if len(self.filter_class) != len(self.filter_content):
            raise ValueError(
                f"Mismatch in number of elements of `filter_class` and `filter_content`: `{self.filter_class}`, `{self.filter_content}`"
            )


def make_parametric_filters_from_json(filename_or_list: Union[str, list]) -> PModel:
    """Make a parametric filter instance from a json representation (from file or list).

    Args:
        filename_or_list (Union[str, list]): JSON filename or list of dictionaries in json format

    Example from file:
    `my_parametric_filters.json`
    ```json
    [
      {
        "filter_class": "network",
        "filter_content": "",
        "pmodel_id": "pricing_range.sample.json"
      },
      {
        "filter_class": "in_period",
        "filter_content": "20220524-20221231",
        "pmodel_id": "pricing_behavior.sample.json"
      },
      {
        "filter_class": "airline",
        "filter_content": "MyAirline",
        "pmodel_id": "pricing_optimizer.sample.json"
      },
      {
        "filter_class": "citysector",
        "filter_content": "CityA->CityB",
        "pmodel_id": "customers_request.poisson_bl1.json"
      },
      {
        "filter_class": "sector",
        "filter_content": "CCCDDD",
        "pmodel_id": "customers_choice.mnl_bl1.json"
      }
    ]
    ```

    ```py
    my_filters = make_parametric_filters_from_json("my_parametric_filters.json")
    ```

    """

    model_id, filename, md5hash, content = parse_data_from_json(filename_or_list)

    if not isinstance(content, list):
        raise ValueError(f"Expected list format in {filename_or_list}")

    content: list

    if len(content) > 0:

        # Used for json validation only, for the moment

        models = list()

        for fields in content:

            filter_class = fields["filter_class"]
            filter_content = fields["filter_content"]

            if isinstance(filter_class, str) and "&" in filter_class:
                filter_class = filter_class.split("&")

            if isinstance(filter_content, str) and "&" in filter_content:
                filter_content = filter_content.split("&")

            models.append(
                PFilter(
                    id=model_id,
                    filter_class=filter_class,
                    filter_content=filter_content,
                    pmodel_id=fields["pmodel_id"],
                )
            )

        return PModel(
            id=model_id, filename=filename, hash=md5hash, content=json.dumps(content)
        )

    else:
        raise ValueError(f"Empty content")



from rmlab._api.upload import APIUploadInternal
from rmlab.data.parametric.filter import PFilter


class APIUploadParametric(APIUploadInternal):
    """Exposes functions for uploading parametric filters to the server."""

    async def upload_parametric_filters(self, scen_id: int, data_fn: str) -> None:
        """Upload a set of parametric filters defined in a file.

        ```json
        {"json": "example"}
        ```

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename (.csv or .json) defining the parametric filters

        Raises:
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
        """

        await self._upload_bounded_items(
            scen_id=scen_id, category=PFilter, data_fn=data_fn
        )



"""
A pricing optimizer completely defines the revenue optimization strategy to apply to 
flights. In particular, it is needed to define:

1. A **schedule**, to define *when* to trigger optimization passes.

2. A **selector**, to specify the *inputs* of the optimization passes for each flight.

3. An **operator**, to specify the set of *actions* and *algorithms* that will run on the input data, encompasing:
"""


from dataclasses import dataclass
import json
from typing import Any, List, Mapping, Optional, Union

from rmlab._data.conversions import parse_data_from_json
from rmlab._data.enums import (
    FileExtensions,
    OptimizationAggregatorKind,
    OptimizationEffectsKind,
    OptimizationForecasterKind,
    OptimizationMaximizerKind,
    OptimizationSelectorFilterKind,
    PricingModelKind,
    TimeUnit,
)
from rmlab.data.parametric._pricing_base import PricingModel


@dataclass
class OptimizerSchedule:
    """Optimizer schedule parameters specify when and
    how often optimization passes are triggered.

    Args:
      from_dbd (int): Positive day before departure from which this schedule is applied.
      day_frequency (int): Number of days between consecutive optimization passes

    Examples:

    * Run every 30 days since it is put on sale:
    ```py
    sch = OptimizerSchedule(from_dbd=float('inf'), day_frequency=30)
    ```

    * Run every 15 days when it reaches dbd 180:
    ```py
    sch = OptimizerSchedule(from_dbd=180, day_frequency=15)
    ```

    * Run every 7 days when it reaches dbd 60:
    ```py
    sch = OptimizerSchedule(from_dbd=60, day_frequency=7)
    ```

    * Run every 3 days when it reaches dbd 30:
    ```py
    sch = OptimizerSchedule(from_dbd=30, day_frequency=3)
    ```

    * Run every day when it reaches dbd 15:
    ```py
    sch = OptimizerSchedule(from_dbd=15, day_frequency=1)
    ```

    """

    from_dbd: int
    day_frequency: int


@dataclass
class OptimizerSelector:
    """Optimizer selector parameters specify which historic data is fed to the forecaster.

    From a functional perspective, a OptimizerSelector instance stores
    a set of parameters that define a function:

    (flightInput) → [flight1, flight2, …, flightN]

    Ie: from the OptimizerSelector parameters we know, given a *flightInput*,
    the set of neighboring flights *[flight1, flight2, …, flightN]* on which
    optimization passes are run upon.

    Args:
      since_qty (int): Time qty
      since_unit (TimeUnit): Time unit
      filters (List[OptimizationSelectorFilterKind]): Filters to apply to input flights

    Examples:

    To create a selector that picks the neighboring flights that:

    * Cover the same citysector and same airline (implicitly assumed, no need to specify it)
    * Depart any time within the last 2 years with respect the flight departure date
    * Departing in the same day of week as flightInput
    * Departing in the same hour slot as flightInput
    * Covering the same sector

    then we create the selector instance as:

    ```py
    sel = OptimizerSelector(
      since_qty=2,
      since_unit="year",
      filters=["day-of-week", "hour-slot", "sector"]
      )
    ```

    See `OptimizationSelectorFilterKind` reference of allowed filters.
    """

    since_qty: int
    since_unit: TimeUnit
    filters: List[OptimizationSelectorFilterKind]

    def __post_init__(self):

        if isinstance(self.since_qty, str) and self.since_qty.lower() == "inf":
            self.since_qty = float("inf")
        else:
            self.since_qty = int(self.since_qty)

        if isinstance(self.since_unit, str):
            self.since_unit = TimeUnit.str_to_enum_value(self.since_unit)

        for i, v in enumerate(self.filters):
            if isinstance(v, str):
                self.filters[i] = OptimizationSelectorFilterKind.str_to_enum_value(v)


class OptimizerMaximizer:
    """Base class for maximizers."""

    kind: OptimizationMaximizerKind


@dataclass
class DPQSDMaximizer(OptimizerMaximizer):
    """Parameters of the DPQSD Maximizer.

    Args:
        max_requests_per_day (int): Upper bound in maximum requests per day
        price_points (int): Number of price points
        ebc_samples (int): Number of samples to take from the optimal solution
    """

    max_requests_per_day: int = 50
    price_points: int = 30
    ebc_samples: int = 5


@dataclass
class OptimizerOperator:
    """Optimizer operator parameters specify the forecasting,
    aggregation and maximization algorithms and which actions are
    performed after an optimization pass finishes.

    Args:
      forecaster_type (OptimizationForecasterKind): Forecaster type
      maximizer_type (OptimizationMaximizerKind): Revenue maximizer type
      aggregate_type (OptimizationAggregatorKind): Type of aggregation of historic input data
      aggregate_value (Optional[float]): Numeric parameter of aggregation type
      effects (str): Actions to perform after maximization

    **Example 1**

    To create an operator that:

    1. Runs *q-forecast* forecaster on all input data
    2. Aggregates forecast data of all flights as a *0.5 exponential" weighting
    3. Modify the thresholds of the flight after maximization pass

    we create an operator instance as:

    ```py
    op = OptimizerOperator(
      forecaster_type="q-forecast",
      maximizer=OptimizerMaximizer(...),
      aggregate_type="exponential",
      aggregate_value=0.5
      effects="commit"
    )
    ```


    **Example 2**

    To create an operator that:

    1. Runs *bayesian* forecaster on all input data
    2. Aggregates forecast data of all flights *uniformly*
    3. Do not apply the results to the thresholds

    we create an operator instance as:

    ```py
    op = OptimizerOperator(
      forecaster_type="q-forecast",
      maximizer=OptimizerMaximizer(...),
      aggregate_type="exponential",
      aggregate_value=0.5
      effects="none"
    )
    ```

    """

    forecaster_type: OptimizationForecasterKind
    maximizer: OptimizerMaximizer
    effects: OptimizationEffectsKind
    aggregate_type: OptimizationAggregatorKind
    aggregate_value: Optional[float] = None

    def __post_init__(self):
        if isinstance(self.forecaster_type, str):
            self.forecaster_type = OptimizationForecasterKind.str_to_enum_value(
                self.forecaster_type
            )

        if not isinstance(self.maximizer, OptimizerMaximizer):
            raise TypeError(f"Unexpected type for maximizer")

        if isinstance(self.aggregate_type, str):
            self.aggregate_type = OptimizationAggregatorKind.str_to_enum_value(
                self.aggregate_type
            )

        if isinstance(self.effects, str):
            self.effects = OptimizationEffectsKind.str_to_enum_value(self.effects)


@dataclass
class OptimizerModel(PricingModel):
    """Dataclass assembling all optimization parameters.

    Args:
      schedule (List[OptimizerSchedule]): Schedule instance
      selector (OptimizerSelector): Selector instance
      operator (OptimizerOperator): Operator instance
      other (Optional[Mapping[str, Any]]): Other fields (for backwards compatibility)
    """

    schedule: List[OptimizerSchedule]
    selector: OptimizerSelector
    operator: OptimizerOperator
    other: Mapping[str, Any] = None

    def __post_init__(self):

        self.schedule = sorted(
            self.schedule, key=lambda sch: sch.from_dbd, reverse=True
        )

        if not isinstance(self.selector, OptimizerSelector):
            raise TypeError(f"Unexpected type for selector: {type(self.selector)}")

        if not isinstance(self.operator, OptimizerOperator):
            raise TypeError(f"Unexpected type for operator: {type(self.operator)}")

        if self.other is None:
            self.other = dict()


def make_pricing_optimizer_from_json(
    filename_or_dict: Union[str, dict]
) -> OptimizerModel:
    """Make a pricing optimizer instance from a json representation (from file or dict).

    Args:
        filename_or_dict (Union[str, dict]): JSON filename or dictionary in json format

    Examples from file:
    ``my_pricing_optimizer.json``
    ```json
    {
      "schedule": {
        "INF": 7,
        "7": 2
      },
      "selector": {
        "since": "2 years",
        "filters": []
      },
      "operator": {
        "forecaster": "q-forecast",
        "aggregate": "exponential 1.5",
        "optimizer": {
          "type": "qsd",
          "max_requests_per_day": 50,
          "price_points": 30,
          "ebc_samples": 5
        },
        "effects": "none"
      }
    }
    ```

    ```py
    my_pricing_optimizer = make_pricing_optimizer_from_json("my_pricing_optimizer.json")
    ```
    """

    model_id, filename, md5hash, content = parse_data_from_json(filename_or_dict)

    if not isinstance(content, dict):
        raise TypeError(f"Expected dict format in {filename_or_dict}")

    content: dict

    mandatory_keys = ["schedule", "selector", "operator"]
    if not all([k in content for k in mandatory_keys]):
        raise ValueError(
            f"Not all mandatory keys present in content. Requred: {mandatory_keys}"
        )

    # ---- Set schedule
    schedule_list = [
        OptimizerSchedule(from_dbd=since, day_frequency=freq)
        for since, freq in content["schedule"].items()
    ]

    # ---- Set selector
    qty, unit = content["selector"]["since"].split(" ")
    filters = content["selector"]["filters"]
    selector = OptimizerSelector(since_qty=qty, since_unit=unit, filters=filters)

    agg_field = content["operator"]["aggregate"]
    if " " in agg_field:
        agg_type, agg_value = agg_field.split(" ")
    else:
        agg_type = agg_field
        agg_value = None

    # ---- Set maximizer
    maximizer_type = content["operator"]["optimizer"]["type"]
    maximizer = None
    if maximizer_type == "qsd":
        maximizer = DPQSDMaximizer(
            max_requests_per_day=content["operator"]["optimizer"][
                "max_requests_per_day"
            ],
            price_points=content["operator"]["optimizer"]["price_points"],
            ebc_samples=content["operator"]["optimizer"]["ebc_samples"],
        )
    else:
        raise ValueError(f"Unknown maximizer type {maximizer_type}")

    other = {ok: ov for ok, ov in content.items() if ok not in mandatory_keys}

    # ---- Set operator
    operator = OptimizerOperator(
        forecaster_type=content["operator"]["forecaster"],
        aggregate_type=agg_type,
        aggregate_value=agg_value,
        maximizer=maximizer,
        effects=content["operator"]["effects"],
    )

    return OptimizerModel(
        id=model_id,
        filename=filename,
        cnt=json.dumps(content),
        hash=md5hash,
        kind=PricingModelKind.OPTIMIZER,
        extension=FileExtensions.JSON,
        schedule=schedule_list,
        selector=selector,
        operator=operator,
        other=other,
    )



"""
This script provides dataclasses storing arrays representing temporal data of flights.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List, Mapping

from rmlab._data.enums import FlightEvent


@dataclass
class FlightData:
    """Base dataclass for derived classes holding time series data associated to a flight.

    Args:
        id (str): Flight ID to which time series data arrays are associated
        timestamps_array (List[datetime]): Sequence of time stamps with milliseconds granularity
    """

    id: str
    timestamps_array: List[datetime]

    def __post_init__(self):
        # NOTE input timestamps are in millisec
        self.timestamps_array = [
            datetime.utcfromtimestamp(ts / 1000) for ts in self.timestamps_array
        ]

    def is_empty(self):
        return len(self.timestamps_array) == 0


@dataclass
class FlightDataBooks(FlightData):
    """Data class with arrays related to books associated to a flight

    Args:
        fares_array (List[str]): Stores the enabled fares at each time stamp
        pps_array (List[int]): Stores the price per seat associated to each booking
        seats_array (List[int]): Stores the seats booked for each booking
        cumulated_seats_array (List[int]): Stores the cumulated booked seats at each time stamp
        cumulated_revenue_array (List[int]): Stores the cumulated revenue at each time stamp
    """

    fares_array: List[str]
    pps_array: List[int]
    seats_array: List[int]
    cumulated_seats_array: List[int]
    cumulated_revenue_array: List[int]


@dataclass
class FlightDataForecastedBooks(FlightData):
    """Data class with arrays related to forecasted books associated to a flight

    Args:
        pps_array (List[int]): Stores the forecasted price per seat at each time stamp
        seats_array (List[int]): Stores the forecasted cumulated booked seats at each time stamp
    """

    pps_array: List[int]
    seats_array: List[int]


@dataclass
class FlightDataThresholdSettings(FlightData):
    """Data class with arrays of seat thresholds for each fare associated to a flight

    Args:
        fare_to_threshold_array (Mapping[str], List[int]): Maps each fare identifier
            to an array storing the seat thresholds at each time stamp.
    """

    fare_to_threshold_array: Mapping[str, List[int]]


@dataclass
class FlightDataPricePerSeatSettings(FlightData):
    """Data class with arrays of prices per seat for each fare associated to a flight

    Args:
        fare_to_pps_array (Mapping[str], List[int]): Maps each fare identifier
            to an array storing the prices per seat at each time stamp.
    """

    fare_to_pps_array: Mapping[str, List[int]]


@dataclass
class FlightDataEvents(FlightData):
    """Data class holding a sequence of events associated to a flight

    Args:
        events_array (List[FlightEvent]): Stores the events at each time stamp.
    """

    events_array: List[FlightEvent]



"""Internal types

Script exposing auxiliary global variables, functions and types to 
provide a minimal typed structure to items.
"""

from dataclasses import dataclass, fields, is_dataclass
from typing import List
import inspect


DateFormat = "%Y-%m-%d"
DateTimeMinFormat = "%Y-%m-%dT%H:%M"
DateTimeSecFormat = "%Y-%m-%dT%H:%M:%S"
DateTimeMicroSecFormat = "%Y-%m-%dT%H:%M:%S.%f"


@dataclass
class APIEvent:
    event: str
    timestamp: int

    def __post_init__(self):
        self.timestamp = int(self.timestamp)


@dataclass
class Item:
    """Base class for all Item classes"""

    id: str


@dataclass
class CustomersModelHolder:
    """Item associated with specific customers models"""

    crequest_id: str
    cchoice_id: str


@dataclass
class PricingModelHolder:
    """Item associated with specific pricing models"""

    prange_id: str
    pbehavior_id: str
    poptimizer_id: str


class CoreItem:
    """Item initialized from direct data upload"""

    pass


class DerivedItem:
    """Item initialized from other item categories"""

    pass


class AirlineLocation:
    """Item that is location attribute of an airline"""

    pass


class BoundedItem(Item):
    """Item whose elements are bounded. The amount of unbounded items grows with the size/space of the underlying scenario"""

    pass


class UnboundedItem(Item):
    """Item whose elements are unbounded. The amount of unbounded items grows with the time period of the underlying scenario, which is in general unlimited"""

    pass


def fields_of_dataclass(category: type) -> List[str]:
    """Return the fields of a dataclass as a list of strings.

    Args:
        category (type): Class of any category

    Returns:
        List of fields as strings
    """

    flds = set()

    for base_class in inspect.getmro(category):
        if is_dataclass(base_class):
            flds.update({f.name for f in fields(base_class)})

    return list(flds)



"""Interface for uploading customers models data."""

import os
from typing import List, Optional
from rmlab._api.upload import APIUploadInternal
from rmlab._data.enums import (
    ParametricModelKind,
    PricingModelKind,
)


class APIUploadPricingModels(APIUploadInternal):
    """Exposes functions for uploading pricing models data to the server."""

    async def upload_pricing_range_model(
        self,
        scen_id: int,
        data_fn: str,
    ) -> None:
        """Upload to server a parametric model defined in file, specifying a pricing range to be applied by flights.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename (.json) defining the parametric model
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.PRICING,
            kind=PricingModelKind.RANGE,
            data_fn=data_fn,
        )

    async def upload_pricing_behavior_model(
        self,
        scen_id: int,
        data_fn: str,
    ) -> None:
        """Upload to server a parametric model defined in file, specifying the pricing behavior/strategy
        under which flights assign prices to seats.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename (.json) defining the parametric model
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.PRICING,
            kind=PricingModelKind.BEHAVIOR,
            data_fn=data_fn,
        )

    async def upload_pricing_optimizer_model(
        self,
        scen_id: int,
        data_fn: str,
    ) -> None:
        """Upload to server a parametric model defined in file, specifying the pricing optimization methods
        under which flights operate, to adapt the prices given current and past demand.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename (.json) defining the parametric model
        """

        await self._upload_parametric_model(
            scen_id,
            parametric_kind=ParametricModelKind.PRICING,
            kind=PricingModelKind.OPTIMIZER,
            data_fn=data_fn,
        )

    async def upload_batch_pricing_models(
        self,
        scen_id: int,
        *,
        range_models_fns: Optional[List[str]] = None,
        behavior_models_fns: Optional[List[str]] = None,
        optimizer_models_fns: Optional[List[str]] = None,
    ):
        if range_models_fns is None:
            range_models_fns = list()
        if behavior_models_fns is None:
            behavior_models_fns = list()
        if optimizer_models_fns is None:
            optimizer_models_fns = list()

        not_existing_fns = (
            [fn for fn in range_models_fns if fn and not os.path.exists(fn)]
            + [fn for fn in behavior_models_fns if fn and not os.path.exists(fn)]
            + [fn for fn in optimizer_models_fns if fn and not os.path.exists(fn)]
        )

        if len(not_existing_fns) > 0:
            raise FileNotFoundError(not_existing_fns)

        for prm in range_models_fns:
            await self._upload_parametric_model(
                scen_id, ParametricModelKind.PRICING, PricingModelKind.RANGE, prm
            )

        for pbm in behavior_models_fns:
            await self._upload_parametric_model(
                scen_id, ParametricModelKind.PRICING, PricingModelKind.BEHAVIOR, pbm
            )

        for pom in optimizer_models_fns:
            await self._upload_parametric_model(
                scen_id, ParametricModelKind.PRICING, PricingModelKind.OPTIMIZER, pom
            )



"""Interface for fetching flight data."""

from typing import List, Optional
from rmlab._data.enums import FlightDataKind
from rmlab._api.fetch import APIFetchInternal

from rmlab.data.flight import (
    FlightDataBooks,
    FlightDataEvents,
    FlightDataPricePerSeatSettings,
    FlightDataThresholdSettings,
)


class APIFetchFlightData(APIFetchInternal):
    """Exposes functions for fetching flight data from the server."""

    async def fetch_flights_data_historic(
        self,
        scen_id: int,
        ids: List[str],
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[FlightDataBooks]:
        """Fetch flight data of historic books of given flights, associated to a citysector or sector, from server.
        At least `citysector_id` and/or `sector_id` associated to the flights must be defined.

        Args:
            scen_id (int): Scenario ID
            ids (List[str]): List of flights IDs
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Target sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined

        Returns:
            List of historic books data containers of each flight.
        """

        return await self._fetch_flights_data(
            scen_id,
            ids,
            FlightDataKind.ACTUAL_BOOKS,
            citysector_id=citysector_id,
            sector_id=sector_id,
        )

    async def fetch_flights_data_pricing_thresholds(
        self,
        scen_id: int,
        ids: List[str],
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[FlightDataThresholdSettings]:
        """Fetch flight data of pricing thresholds of given flights, associated to a citysector or sector, from server.
        At least `citysector_id` and/or `sector_id` associated to the flights must be defined.

        Args:
            scen_id (int): Scenario ID
            ids (List[str]): List of flights IDs
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Target sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined

        Returns:
            List of pricing thresholds data containers of each flight.
        """

        return await self._fetch_flights_data(
            scen_id,
            ids,
            FlightDataKind.THRESHOLDS_SETTINGS,
            citysector_id=citysector_id,
            sector_id=sector_id,
        )

    async def fetch_flights_data_pricing_per_seat(
        self,
        scen_id: int,
        ids: List[str],
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[FlightDataPricePerSeatSettings]:
        """Fetch flight data of pricing per seat of given flights, associated to a citysector or sector, from server.
        At least `citysector_id` and/or `sector_id` associated to the flights must be defined.

        Args:
            scen_id (int): Scenario ID
            ids (List[str]): List of flights IDs
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Target sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined

        Returns:
            List of pricing per seat data containers of each flight.
        """

        return await self._fetch_flights_data(
            scen_id,
            ids,
            FlightDataKind.PRICE_PER_SEAT_SETTINGS,
            citysector_id=citysector_id,
            sector_id=sector_id,
        )

    async def fetch_flights_data_events(
        self,
        scen_id: int,
        ids: List[str],
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[FlightDataEvents]:
        """Fetch flight data of events of given flights, associated to a citysector or sector, from server.
        At least `citysector_id` and/or `sector_id` associated to the flights must be defined.

        Args:
            scen_id (int): Scenario ID
            ids (List[str]): List of flights IDs
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Target sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined

        Returns:
            List of events data containers of each flight.
        """

        return await self._fetch_flights_data(
            scen_id,
            ids,
            FlightDataKind.EVENTS,
            citysector_id=citysector_id,
            sector_id=sector_id,
        )



"""Interface for remote data removal

This script provides a APIRemove context manager exposing functions 
for removing data from the server.
"""

from rmlab._api.remove import APIRemoveInternal
from rmlab._data.enums import DataRemoveKind


class APIRemove(APIRemoveInternal):
    """Interface to upload data to server"""

    async def remove_data_full(self, scen_id: int) -> None:
        """Remove all the data of a scenario.

        Args:
            scen_id (int): Scenario ID
        """

        await self._remove_data(scen_id, DataRemoveKind.COMPLETE)

    async def remove_data_historic(self, scen_id: int) -> None:
        """Remove all the historic flight data of a scenario.

        Args:
            scen_id (int): Scenario ID
        """

        await self._remove_data(scen_id, DataRemoveKind.HISTORIC)



from typing import Any, List, Mapping, Optional, Tuple

from rmlab._data.enums import FlightDataKind, ScenarioState
from rmlab._data.types import (
    AirlineLocation,
    BoundedItem,
    UnboundedItem,
    fields_of_dataclass,
)
from rmlab_http_client import Cache

from rmlab._api.base import APIBaseInternal
from rmlab._data.conversions import S2C_DataKind2ConvertFunction
from rmlab.data.scenario import (
    ItemsCount,
    FlightsCount,
    ScenarioDates,
    SchedulesCount,
)
from rmlab.data.items import (
    AirlineLocationItems,
    AllBoundedItems,
    AnyAirlineLocationType,
    AnyBoundedType,
    AnyUnboundedType,
    Flight,
    Schedule,
)
from rmlab.data.flight import FlightData


class APIFetchInternal(APIBaseInternal):
    """Interface to fetch data from server."""

    async def _fetch_info(
        self, scen_id: int
    ) -> Tuple[ScenarioDates, ItemsCount, SchedulesCount, FlightsCount]:
        """Fetch summarized information about a scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            Summary information about a scenario.
        """

        summary = await self._submit_call(
            "api-data-summary-get-single", scen_id=scen_id
        )

        return self._process_summary(scen_id, summary)

    def _process_summary(
        self, scen_id: int, summary: Mapping[str, Any]
    ) -> Tuple[ScenarioDates, ItemsCount, SchedulesCount, FlightsCount]:

        return (
            ScenarioDates(
                id=scen_id,
                upload_time=summary["upload_time"],
                checkpoints=summary["checkpoints"],
                state=ScenarioState.str_to_enum_value(summary["state"]),
                **summary["dates"],
            ),
            ItemsCount(
                id=scen_id,
                **{
                    bi.__name__.lower(): summary["itemsCount"][bi.__name__.lower()]
                    for bi in AllBoundedItems
                },
            ),
            SchedulesCount(
                id=scen_id,
                **summary["itemsCount"][Schedule.__name__.lower()],
            ),
            FlightsCount(
                id=scen_id,
                **summary["itemsCount"][Flight.__name__.lower()],
            ),
        )

    async def _fetch_bounded_items(
        self, scen_id: int, category: AnyBoundedType
    ) -> List[BoundedItem]:
        """Fetch a list of items of bounded category from server.

        Args:
            scen_id (int): Scenario ID
            category (AnyBoundedType): Category of items

        Raises:
            ValueError: If category is not bounded

        Returns:
            List[BoundedItem]: List of bounded items
        """

        if not issubclass(category, BoundedItem):
            raise ValueError(
                f"Category `{category.__name__}` must inherit from `BoundedItem`"
            )

        resp_data = await self._submit_call(
            "api-data-bounded-get-all",
            scen_id=scen_id,
            category=category.__name__.lower(),
        )

        return [category(**itm) for itm in resp_data["items"]]

    async def _fetch_airline_locations_ids(
        self, scen_id: int, airline_id: str, location_type: AnyAirlineLocationType
    ) -> List[str]:
        """Fetch a list of location identifiers associated to an airline from server.
        Used To fetch identifiers of all sectors / citysectors / routes / cityroutes covered by an airline.
        Eg: await api._fetch_airline_locations_ids(scen_id=0, airline_id='MyAirline', location_type=rmlab.data.types.Sector)

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID
            location_type (AnyAirlineLocationType): Class type, any of `Sector`, `Citysector`, `Route`, `CityRoute`, of `rmlab.data.types` module

        Raises:
            ValueError: If location type is invalid

        Returns:
            List[str]: List of location identifiers
        """

        if location_type not in AirlineLocationItems:
            raise ValueError(
                f"Location type must be one of `{AirlineLocationItems}`, got `{location_type}`"
            )

        return await self._submit_call(
            "api-data-airline-locations-get-ids",
            scen_id=scen_id,
            airline_id=airline_id,
            location_type=location_type.__name__.lower(),
        )

    async def _fetch_airline_locations_items(
        self, scen_id: int, airline_id: str, location_type: AnyAirlineLocationType
    ) -> List[AirlineLocation]:
        """Fetch a list of location items associated to an airline from server.
        Used To fetch items of all sectors / citysectors / routes / cityroutes covered by an airline.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID
            location_type (AnyAirlineLocationType): Class type, any of `Sector`, `Citysector`, `Route`, `CityRoute`, of `rmlab.data.types` module

        Raises:
            ValueError: If location type is invalid

        Returns:
            List[AirlineLocation]: List of location items
        """

        if location_type not in AirlineLocationItems:
            raise ValueError(
                f"Location type must be one of `{AirlineLocationItems}`, got `{location_type}`"
            )

        items = await self._submit_call(
            "api-data-airline-locations-get-items",
            scen_id=scen_id,
            airline_id=airline_id,
            location_type=location_type.__name__.lower(),
        )

        return [location_type(**item) for item in items]

    async def _fetch_unbounded_ids(
        self,
        scen_id: int,
        category: AnyUnboundedType,
        *,
        citysector_id: Optional[str] = None,
        airline_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[str]:
        """Fetch all identifiers of an unbounded category associated to a citysector, sector (and/or) airline from server.
        At least `citysector_id` and/or `sector_id` must be defined.

        Args:
            scen_id (int): Scenario ID
            category (AnyUnboundedType): Unbounded category (ie: `rmlab.api.data.types.Schedule` or `rmlab.api.data.types.Flight`)
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None
            airline_id (Optional[str], optional): Target airline ID. Defaults to None
            sector_id (Optional[str], optional): Target sector ID. Defaults to None

        Raises:
            ValueError: If category is not unbounded
            ValueError: If none of `citysector_id`, `airline_id`, `sector_id` is defined
            ValueError: If only `airline_id` is defined

        Returns:
            List[str]: List of identifiers of unbounded items.
        """

        if not issubclass(category, UnboundedItem):
            raise ValueError(
                f"Category `{category.__name__}` must inherit from `UnboundedItem`"
            )

        if airline_id is None and citysector_id is None and sector_id is None:
            raise ValueError(
                f"Require definition of `citysector_id`, `airline_id` or `sector_id`"
            )

        if airline_id is not None and (citysector_id is None and sector_id is None):
            raise ValueError(
                f"Required definition of `citysector_id` and/or `sector_id` when airline_id is defined"
            )

        return await self._submit_call(
            "api-data-unbounded-get-ids",
            scen_id=scen_id,
            category=category.__name__.lower(),
            citysector_id=citysector_id,
            airline_id=airline_id,
            sector_id=sector_id,
        )

    async def _fetch_unbounded_items(
        self,
        scen_id: int,
        category: AnyUnboundedType,
        ids: List[str],
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[UnboundedItem]:
        """Fetch all items of an unbounded category associated to a citysector or sector from server.
        At least `citysector_id` and/or `sector_id` associated to the items must be defined.

        Args:
            scen_id (int): Scenario ID
            category (AnyUnboundedType): Unbounded category (ie: `rmlab.api.data.types.Schedule` or `rmlab.api.data.types.Flight`)
            ids (List[str]): List of flights schedules or flights IDs
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None
            sector_id (Optional[str], optional): Target sector ID. Defaults to None

        Raises:
            ValueError: If category is not unbounded
            ValueError: If none of `citysector_id`, `sector_id` is defined

        Returns:
            List[UnboundedItem]: List of items of unbounded category.
        """

        if not issubclass(category, UnboundedItem):
            raise ValueError(
                f"Category `{category.__name__}` must inherit from `UnboundedItem`"
            )

        if citysector_id is None and sector_id is None:
            raise ValueError(
                f"At least one of `citysector_id`, `sector_id` must be defined"
            )

        fields = fields_of_dataclass(category)

        endpoint_id = "api-data-unbounded-get-fields"
        ep = Cache.get_endpoint(endpoint_id)

        endpoint_limit = ep.arguments.limit_of("ids")
        if endpoint_limit is None:
            endpoint_limit = float("inf")

        if len(ids) > endpoint_limit:

            # Sequentially fetch items in chunks to keep the request/response payloads and server load controlled

            ret_fields: List[UnboundedItem] = list()

            for chunk_start in range(0, len(ids), endpoint_limit):

                chunk_end = chunk_start + endpoint_limit
                if chunk_end > len(ids):
                    chunk_end = -1

                raw_list = await self._submit_call(
                    endpoint_id,
                    scen_id=scen_id,
                    category=category.__name__.lower(),
                    citysector_id=citysector_id,
                    sector_id=sector_id,
                    ids=ids[chunk_start:chunk_end],
                    fields=fields,
                )

                ret_fields += [category(**fields) for fields in raw_list]

            return ret_fields

        else:

            raw_list = await self._submit_call(
                endpoint_id,
                scen_id=scen_id,
                category=category.__name__.lower(),
                citysector_id=citysector_id,
                sector_id=sector_id,
                ids=ids,
                fields=fields,
            )

            return [category(**fields) for fields in raw_list]

    async def _fetch_flights_data(
        self,
        scen_id: int,
        ids: List[str],
        kind: FlightDataKind,
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[FlightData]:
        """Fetch flight data arrays of given flights, associated to a citysector or sector, from server.
        At least `citysector_id` and/or `sector_id` associated to the flights must be defined.

        Args:
            scen_id (int): Scenario ID
            ids (List[str]): List of flights IDs
            kind (FlightDataKind): Kind of flight data to fetch
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Target sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined

        Returns:
            List[FlightData]: List of containers holding the flight data arrays.
        """

        if citysector_id is None and sector_id is None:
            raise ValueError(
                f"Require definition of either `citysector_id` or `sector_id`"
            )

        endpoint_id = "api-data-flight-get"
        ep = Cache.get_endpoint(endpoint_id)

        endpoint_limit = ep.arguments.limit_of("ids")
        if endpoint_limit is None:
            endpoint_limit = float("inf")

        if len(ids) > endpoint_limit:

            # Sequentially fetch items in chunks to keep the request/response payloads and server load controlled

            ret_data: List[FlightData] = list()

            for chunk_start in range(0, len(ids), endpoint_limit):

                chunk_end = chunk_start + endpoint_limit
                if chunk_end > len(ids):
                    chunk_end = -1

                flights_data = await self._submit_call(
                    endpoint_id,
                    scen_id=scen_id,
                    ids=ids[chunk_start:chunk_end],
                    kind=kind.value,
                    citysector_id=citysector_id,
                    sector_id=sector_id,
                )

                ret_data += [
                    S2C_DataKind2ConvertFunction[kind](flight_id, flight_data)
                    for flight_id, flight_data in zip(
                        ids[chunk_start:chunk_end], flights_data
                    )
                ]

            return ret_data

        else:

            flights_data = await self._submit_call(
                endpoint_id,
                scen_id=scen_id,
                ids=ids,
                kind=kind.value,
                citysector_id=citysector_id,
                sector_id=sector_id,
            )

            return [
                S2C_DataKind2ConvertFunction[kind](flight_id, flight_data)
                for flight_id, flight_data in zip(ids, flights_data)
            ]



"""Interface for core items data fetching."""

from typing import List, Optional, Tuple
from rmlab.data.items import (
    Aircraft,
    Airline,
    Airport,
    City,
    CityRoute,
    CitySector,
    Country,
    Flight,
    Route,
    Schedule,
    Sector,
)


from rmlab._api.fetch import APIFetchInternal
from rmlab.data.scenario import (
    ItemsCount,
    FlightsCount,
    ScenarioDates,
    SchedulesCount,
)


class APIFetchCore(APIFetchInternal):
    """Exposes functions for fetching core items data from the server."""

    async def fetch_info(
        self, scen_id: int
    ) -> Tuple[ScenarioDates, ItemsCount, SchedulesCount, FlightsCount]:
        """Fetch summarized information about a scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            A 4-tuple with information about the scenario.
        """

        return await self._fetch_info(scen_id)

    async def fetch_aircrafts(self, scen_id: int) -> List[Aircraft]:
        """Fetch a list of all aircrafts of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of aircrafts
        """

        return await self._fetch_bounded_items(scen_id, Aircraft)

    async def fetch_airlines(self, scen_id: int) -> List[Airline]:
        """Fetch a list of all airlines of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of airlines
        """

        return await self._fetch_bounded_items(scen_id, Airline)

    async def fetch_airports(self, scen_id: int) -> List[Airport]:
        """Fetch a list of all airport of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of airport
        """

        return await self._fetch_bounded_items(scen_id, Airport)

    async def fetch_countries(self, scen_id: int) -> List[Country]:
        """Fetch a list of all countries of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of countries
        """

        return await self._fetch_bounded_items(scen_id, Country)

    async def fetch_cities(self, scen_id: int) -> List[City]:
        """Fetch a list of all cities of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of cities
        """

        return await self._fetch_bounded_items(scen_id, City)

    async def fetch_citysectors(self, scen_id: int) -> List[CitySector]:
        """Fetch a list of all citysectors of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of citysectors
        """

        return await self._fetch_bounded_items(scen_id, CitySector)

    async def fetch_sectors(self, scen_id: int) -> List[Sector]:
        """Fetch a list of all sectors of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of sectors
        """

        return await self._fetch_bounded_items(scen_id, Sector)

    async def fetch_cityroutes(self, scen_id: int) -> List[CityRoute]:
        """Fetch a list of all cityroutes of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of cityroutes
        """

        return await self._fetch_bounded_items(scen_id, CityRoute)

    async def fetch_routes(self, scen_id: int) -> List[Route]:
        """Fetch a list of all routes of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of routes
        """

        return await self._fetch_bounded_items(scen_id, Route)

    async def fetch_airline_sectors(
        self, scen_id: int, airline_id: str
    ) -> List[Sector]:
        """Fetch a list of sectors associated to an airline from server.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID

        Returns:
            List of sectors
        """

        return await self._fetch_airline_locations_items(scen_id, airline_id, Sector)

    async def fetch_airline_citysectors(
        self, scen_id: int, airline_id: str
    ) -> List[CitySector]:
        """Fetch a list of citysectors associated to an airline from server.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID

        Returns:
            List of citysectors
        """

        return await self._fetch_airline_locations_items(
            scen_id, airline_id, CitySector
        )

    async def fetch_airline_routes(self, scen_id: int, airline_id: str) -> List[Route]:
        """Fetch a list of routes associated to an airline from server.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID

        Returns:
            List of routes
        """

        return await self._fetch_airline_locations_items(scen_id, airline_id, Route)

    async def fetch_airline_cityroutes(
        self, scen_id: int, airline_id: str
    ) -> List[CityRoute]:
        """Fetch a list of cityroutes associated to an airline from server.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID

        Returns:
            List of cityroutes
        """

        return await self._fetch_airline_locations_items(scen_id, airline_id, CityRoute)

    async def fetch_airline_sectors_ids(
        self, scen_id: int, airline_id: str
    ) -> List[str]:
        """Fetch a list of sectors ids associated to an airline from server.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID

        Returns:
            List of sectors
        """

        return await self._fetch_airline_locations_ids(scen_id, airline_id, Sector)

    async def fetch_airline_citysectors_ids(
        self, scen_id: int, airline_id: str
    ) -> List[str]:
        """Fetch a list of citysectors ids associated to an airline from server.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID

        Returns:
            List of citysectors
        """

        return await self._fetch_airline_locations_ids(scen_id, airline_id, CitySector)

    async def fetch_airline_routes_ids(
        self, scen_id: int, airline_id: str
    ) -> List[str]:
        """Fetch a list of routes ids associated to an airline from server.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID

        Returns:
            List of routes
        """

        return await self._fetch_airline_locations_ids(scen_id, airline_id, Route)

    async def fetch_airline_cityroutes_ids(
        self, scen_id: int, airline_id: str
    ) -> List[str]:
        """Fetch a list of cityroutes associated to an airline from server.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID

        Returns:
            List of cityroutes
        """

        return await self._fetch_airline_locations_ids(scen_id, airline_id, CityRoute)

    async def fetch_schedules_ids(
        self,
        scen_id: int,
        *,
        citysector_id: Optional[str] = None,
        airline_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[str]:
        """Fetch all flights schedules identifiers associated to a citysector or sector from server.
        At least `citysector_id` and/or `sector_id` associated to the items must be defined.

        Args:
            scen_id (int): Scenario ID
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None
            airline_id (Optional[str], optional): Target airline ID. Defaults to None
            sector_id (Optional[str], optional): Target sector ID. Defaults to None

        Raises:
            ValueError: If none of `citysector_id`, `airline_id`, `sector_id` is defined
            ValueError: If only `airline_id` is defined

        Returns:
            List of identifiers of unbounded items
        """

        return await self._fetch_unbounded_ids(
            scen_id,
            Schedule,
            citysector_id=citysector_id,
            airline_id=airline_id,
            sector_id=sector_id,
        )

    async def fetch_schedules(
        self,
        scen_id: int,
        ids: List[str],
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[Schedule]:
        """Fetch all flights schedules associated to a citysector or sector from server.
        At least `citysector_id` and/or `sector_id` associated to the items must be defined.

        Args:
            scen_id (int): Scenario ID
            ids (List[str]): List of schedules IDs
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None
            sector_id (Optional[str], optional): Target sector ID. Defaults to None

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined

        Returns:
            List of flights schedules.
        """

        return await self._fetch_unbounded_items(
            scen_id, Schedule, ids, citysector_id=citysector_id, sector_id=sector_id
        )

    async def fetch_flights_ids(
        self,
        scen_id: int,
        *,
        citysector_id: Optional[str] = None,
        airline_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[str]:
        """Fetch all flights identifiers associated to a citysector or sector from server.
        At least `citysector_id` and/or `sector_id` associated to the items must be defined.

        Args:
            scen_id (int): Scenario ID
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None
            airline_id (Optional[str], optional): Target airline ID. Defaults to None
            sector_id (Optional[str], optional): Target sector ID. Defaults to None

        Raises:
            ValueError: If none of `citysector_id`, `airline_id`, `sector_id` is defined
            ValueError: If only `airline_id` is defined

        Returns:
            List of identifiers of unbounded items
        """

        return await self._fetch_unbounded_ids(
            scen_id,
            Flight,
            citysector_id=citysector_id,
            airline_id=airline_id,
            sector_id=sector_id,
        )

    async def fetch_flights(
        self,
        scen_id: int,
        ids: List[str],
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> List[Flight]:
        """Fetch all flights associated to a citysector or sector from server.
        At least `citysector_id` and/or `sector_id` associated to the items must be defined.

        Args:
            scen_id (int): Scenario ID
            ids (List[str]): List of flights IDs
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None
            sector_id (Optional[str], optional): Target sector ID. Defaults to None

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined

        Returns:
            List of flights
        """

        return await self._fetch_unbounded_items(
            scen_id, Flight, ids, citysector_id=citysector_id, sector_id=sector_id
        )



"""Interface for running simulations."""

import asyncio
from datetime import datetime
from typing import List, Optional, Tuple

from rmlab._data.types import DateFormat
from rmlab._api.simulation import APISimulationInternal
from rmlab.data.scenario import (
    ItemsCount,
    FlightsCount,
    ScenarioDates,
    SchedulesCount,
)


class APISimulation(APISimulationInternal):
    """Exposes functions for running simulations on server and adding simulation checkpoints."""

    async def upload_checkpoints(
        self, scen_id: int, checkpoints: List[datetime]
    ) -> None:
        """Upload date checkpoints at which simulation is paused.

        Args:
            scen_id (int): Scenario ID in which the simulation is running
            checkpoints (List[datetime]): List of checkpoints
        """

        await self._submit_call(
            "api-operation-simulation-checkpoint",
            scen_id=scen_id,
            checkpoints=[datetime.strftime(chp, DateFormat) for chp in checkpoints],
        )

    async def trigger_simulation(
        self, scen_id: int, next: Optional[datetime] = None
    ) -> Tuple[ScenarioDates, ItemsCount, SchedulesCount, FlightsCount]:
        """Trigger a simulation run on a given scenario.

        Args:
            scen_id (int): Scenario ID
            next (Optional[datetime], optional): Checkpoint at which the simulation is stopped. Defaults to None.

        Raises:
            ValueError: If the type of `next` is invalid
            RuntimeError: If simulation failed for any reason

        Returns:
            Summarized information of the scenario after the simulation finishes
        """

        if next is not None:
            if not isinstance(next, datetime):
                raise ValueError(
                    f"Expected `datetime` type in `next` {next}, got `{type(next)}`"
                )

            await self._submit_call(
                "api-operation-simulation-checkpoint",
                scen_id=scen_id,
                checkpoints=[datetime.strftime(next, DateFormat)],
            )

        trigger_task = asyncio.create_task(
            self._submit_call(
                "api-operation-simulation-trigger", scen_id=scen_id, operation="auto"
            )
        )

        stop_event = asyncio.Event()

        progress_task = asyncio.create_task(
            self._periodic_trigger_progress(scen_id, stop_event)
        )

        await asyncio.wait(
            [trigger_task, progress_task],
            timeout=None,
            return_when=asyncio.FIRST_COMPLETED,
        )

        if not stop_event.is_set() and not progress_task.done():
            stop_event.set()
            await progress_task

        exc = progress_task.exception()
        if exc is not None:
            raise RuntimeError(f"Error {type(exc)} while tracking progress: {exc}")

        summary = trigger_task.result()

        return self._process_summary(scen_id, summary)



"""Provides utility functions."""

import asyncio
from typing import Any, Coroutine, List, Union
from rmlab_errors import MultipleError


async def run_async_tasks(
    tasks: List[Coroutine], return_results: bool = False, discard_none: bool = False
) -> Union[List[Any], None]:
    """Runs multiple asynchronous tasks concurrently capturing exceptions.

    Args:
        tasks (List[Coroutine]): List of tasks to run concurrently.
        return_results (bool, optional): Whether to return results of tasks. Defaults to False.
        discard_none (bool, optional): If returning results, whether to discard the tasks returning ``None``. Defaults to False.

    Raises:
        MultipleError: If more than one task raised an exception.
        Exception: If a single task raised an exception.

    Returns:
        Union[List[Any], NoneType]: If ``return_results`` is ``True``, None otherwise.
    """

    results = await asyncio.gather(*tasks, return_exceptions=True)

    errors = [res for res in results if isinstance(res, Exception)]
    if len(errors) > 1:
        raise MultipleError(*errors)
    elif len(errors) == 1:
        raise errors[0]

    if return_results:
        if discard_none:
            return [r for r in results if r is not None]
        else:
            return list(results)



from dataclasses import dataclass
from rmlab._data.enums import CustomersModelKind, FileExtensions
from rmlab.data.items import PModel


@dataclass
class CustomersModel(PModel):
    """Customers models docstring"""

    kind: CustomersModelKind
    extension: FileExtensions



from rmlab._api.base import APIBaseInternal
from rmlab._data.enums import (
    DataRemoveKind,
)


class APIRemoveInternal(APIBaseInternal):
    """Interface to remove data from server."""

    async def _remove_data(self, scen_id: int, kind: DataRemoveKind) -> None:
        """Remove some or all the data of a scenario.

        Args:
            scen_id (int): Scenario ID.
            kind (DataRemoveKind): Type of removal.
        """

        if kind.value == "full":

            await self._submit_call("api-data-remove-full", scen_id=scen_id)

        elif kind.value == "restart":

            await self._submit_call("api-data-remove-restart", scen_id=scen_id)

        else:

            raise ValueError(f"Unknown kind `{kind}`")



"""Script exposing the core underlying API functions, handling authentication and API calls."""

import os, logging
from types import MethodType
from typing import Any, Iterable, List, Mapping, Optional

from rmlab_http_client import (
    AsyncClient,
    AuthType,
    Endpoint,
    PayloadArguments,
    ResponseType,
    SyncClient,
    MethodType,
    CommunicationType,
    PayloadType,
    DataRequestContext,
    DataRequestContextMultipart,
    Cache,
)

# Endpoint to discover dynamic protected auth endpoints
DiscoverAuthEndpoint = Endpoint(
    resource=["auth", "discover"],
    method=MethodType.GET,
    payload=PayloadType.NONE,
    arguments=PayloadArguments(),
    auth=AuthType.PUBLIC,
    response=ResponseType.JSON,
    id="auth-discover",
)

from rmlab._version import __version__

_Logger = logging.getLogger(__name__)

BaseURL = "https://rmlab.ai"

_ExpectedCredentialsKeys = [
    "access_token",
    "refresh_token",
    "services",
    "scenarios",
    "username",
    "workgroup",
]

_AuthEndpoints = [
    "auth-log-in",
    "auth-log-in-poll-status",
    "auth-log-in-result",
    "auth-log-out",
    "auth-token-refresh",
]

_ApiEndpoints = [
    "api-discover-user",
    "api-data-bounded-get-meta",
    "api-data-bounded-get-all",
    "api-data-bounded-get-single",
    "api-data-bounded-post",
    "api-data-flight-get",
    "api-data-flight-post",
    "api-data-flight-thresholds-post",
    "api-data-pmodel-post",
    "api-data-unbounded-get-fields",
    "api-data-unbounded-get-ids",
    "api-data-unbounded-post",
    "api-data-airline-locations-get-ids",
    "api-data-airline-locations-get-items",
    "api-data-remove-full",
    "api-data-remove-restart",
    "api-data-summary-get-all",
    "api-data-summary-get-single",
    "api-operation-simulation-checkpoint",
    "api-operation-simulation-pause",
    "api-operation-simulation-trigger",
    "api-operation-optimization-trigger",
    "api-operation-optimization-schedule",
    "api-monitor-activity-user",
    "api-monitor-async-scenario-status",
    "api-monitor-async-scenario-result",
]

_ExpectedEndpointsIds = _AuthEndpoints + _ApiEndpoints


def _check_valid_credentials(creds: Mapping[str, Any]):

    if not all([k in creds for k in _ExpectedCredentialsKeys]):
        raise ValueError(f"Some required keys not found in received credentials")


def _check_valid_endpoints(endpoints_ids: Iterable[str]):

    for ep_id in endpoints_ids:
        if ep_id not in _ExpectedEndpointsIds:
            _Logger.warning(f"Unrecognized endpoint `{ep_id}`")


class APIBaseInternal:
    """Asynchronous context manager providing internal initialization, login and API server calls to derived classes."""

    def __init__(
        self,
        workgroup: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        app_url: Optional[str] = None,
        admin_url: Optional[str] = None
    ) -> None:
        """Initializes APIBaseInternal instance.

        Args:
            workgroup: (str, optional): Workgroup name. Defaults to None.
                If none, expects `RMLAB_WORKGROUP` environment variable definition.
            user (str, optional): User name. Defaults to None.
                If none, expects `RMLAB_USERNAME` environment variable definition.
            password (str, optional): Password. Defaults to None.
                If none, expects `RMLAB_PASSWORD` environment variable definition..

        Raises:
            AuthenticationError: If user or password are not defined.
        """

        if any([arg is None for arg in [workgroup, username, password]]):

            try:
                workgroup = os.environ["RMLAB_WORKGROUP"]
                username = os.environ["RMLAB_USERNAME"]
                password = os.environ["RMLAB_PASSWORD"]

            except KeyError:

                raise ValueError(
                    f"`RMLAB_WORKGROUP` / `RMLAB_USERNAME` / `RMLAB_PASSWORD` are undefined"
                )

        Cache.set_credential(
            "basic_auth", ":".join([workgroup, username, password, __version__])
        )

        self._url_api = app_url or BaseURL
        self._url_auth = admin_url or BaseURL

    @classmethod
    @property
    def scenarios(cls) -> List[int]:
        """Return the scenarios integer identifiers.

        Raises:
            ValueError: If scenarios are not initialized

        Returns:
            List[int]: Ordered list of integer scenarios ids.
        """
        scenarios = Cache.get_credential("scenarios")
        if scenarios:
            scenarios_keys = list(scenarios.keys())
            if all([isinstance(k, str) for k in scenarios_keys]):
                scenarios_keys.sort()
                scenarios_int_keys = {k: scenarios[str(k)] for k in scenarios_keys}
                Cache.set_credential("scenarios", scenarios_int_keys)
            return scenarios_keys
        else:
            raise ValueError(f"Scenarios not initialized")

    async def __aenter__(self):
        """Context manager asynchronous enter.
        If credentials are not defined:
        1. Fetch and cache dynamic authentication endpoints
        2. Uses those endpoints to log-in, and cache credentials
        3. Fetch and cache dynamic api endpoints

        Returns:
            APIBaseInternal: This instance.
        """

        if Cache.get_credential("access_token") is None:

            async with SyncClient(
                DiscoverAuthEndpoint, address=self._url_auth
            ) as discover_auth_client:

                auth_endpoints = await discover_auth_client.submit_request()
                _check_valid_endpoints(auth_endpoints.keys())
                Cache.add_endpoints(**auth_endpoints)

            async with AsyncClient(
                address=self._url_auth,
                async_endpoint=Cache.get_endpoint("auth-log-in"),
                poll_endpoint=Cache.get_endpoint("auth-log-in-poll-status"),
                result_endpoint=Cache.get_endpoint("auth-log-in-result"),
            ) as login_client:

                _Logger.debug("Logging-in...")

                creds = await login_client.submit_request()
                _check_valid_credentials(creds)
                for ck, cv in creds.items():
                    Cache.set_credential(ck, cv)

                endpoints: Mapping[str, dict] = creds.pop("endpoints")
                _check_valid_endpoints(endpoints.keys())
                Cache.add_endpoints(**endpoints)

            async with SyncClient(
                Cache.get_endpoint("api-discover-user"),
                address=self._url_api,
                refresh_address=self._url_auth,
                refresh_endpoint=Cache.get_endpoint("auth-token-refresh"),
            ) as discover_api_client:

                api_endpoints = await discover_api_client.submit_request()
                _check_valid_endpoints(api_endpoints.keys())
                Cache.add_endpoints(**api_endpoints)

        return self

    async def __aexit__(self, exc_ty, exc_val, exc_tb):
        """Context manager asynchronous exit.
        Submits log-out to server, but does not re-raise any eventual error"""

        try:

            async with SyncClient(
                Cache.get_endpoint("auth-log-out"), address=self._url_auth
            ) as logout_client:

                await logout_client.submit_request()

        except Exception as exc:
            # Do not propagate exceptions due to log-out
            _Logger.warning(f"Absorbing log-out exception {exc}")
            pass

    async def _submit_call(self, endpoint_id: str, **kwargs) -> Any:
        """Submit API call to a specific server endpoint.

        Args:
            endpoint_id (str): Endpoint identifier

        Raises:
            RuntimeError: If endpoint does not exist
            ValueError: If arguments do not match endpoint requirements

        Returns:
            Any: API response
        """

        ep = Cache.get_endpoint(endpoint_id)

        context_type = (
            DataRequestContextMultipart
            if ep.payload == PayloadType.MULTIPART
            else DataRequestContext
        )

        with context_type(ep, **kwargs) as rdc:

            if ep.communication == CommunicationType.SYNC:

                async with SyncClient(
                    ep,
                    address=self._url_api,
                    refresh_address=self._url_auth,
                    refresh_endpoint=Cache.get_endpoint("auth-token-refresh"),
                ) as sync_client:

                    return await sync_client.submit_request(rdc.data)

            else:

                assert ep.communication == CommunicationType.ASYNC

                async with AsyncClient(
                    address=self._url_api,
                    async_endpoint=ep,
                    poll_endpoint=Cache.get_endpoint(ep.async_poll_endpoint_id),
                    result_endpoint=Cache.get_endpoint(ep.async_result_endpoint_id),
                    refresh_address=self._url_auth,
                    refresh_endpoint=Cache.get_endpoint("auth-token-refresh"),
                ) as async_client:

                    return await async_client.submit_request(rdc.data)



"""
This script provides dataclasses that represent the relational data components in the server.
"""

from typing import List, Union
from dataclasses import dataclass
from datetime import datetime, timedelta
import importlib
import inspect
from rmlab._data.enums import CurrencyKind, DayOfWeek

from rmlab._data.types import (
    AirlineLocation,
    BoundedItem,
    CoreItem,
    CustomersModelHolder,
    DateFormat,
    DateTimeMinFormat,
    DerivedItem,
    PricingModelHolder,
    UnboundedItem,
)


@dataclass
class PModel(BoundedItem, CoreItem):
    """Item holding a set of parameters.

    Args:
        filename (str): Name of file storing the parameters
        hash (str): MD5 hash of the content
        cnt (str): Model content
    """

    filename: str
    hash: str
    cnt: str


@dataclass
class Airline(BoundedItem, CoreItem):
    """Airline instance.

    Args:
        name (str): Airline name and ID
        type (str): Airline type (Either 'low-cost' or 'legacy')
    """

    name: str
    type: str = "undefined"


@dataclass
class Aircraft(BoundedItem, CoreItem):
    """Aircraft instance.

    Args:
        model (str): Aircraft model and ID
        seat_capacity (int): Seat capacity
    """

    model: str
    seat_capacity: int

    def __post_init__(self):
        self.seat_capacity = int(self.seat_capacity)


@dataclass
class Country(BoundedItem, CoreItem):
    """Country instance.

    Args:
        name (str): Country name and ID
        currency (CurrencyKind): Country currency
    """

    name: str
    currency: CurrencyKind

    def __post_init__(self):
        if isinstance(self.currency, str):
            self.currency = CurrencyKind.str_to_enum_value(self.currency)


@dataclass
class City(BoundedItem, CoreItem):
    """City instance.

    Args:
        name (str): City name and ID
        country_id (str): Country ID of the city
    """

    name: str
    country_id: str


@dataclass
class CitySector(BoundedItem, DerivedItem, CustomersModelHolder, AirlineLocation):
    """Pair of cities with specified direction.

    Args:
        origin_id (str): Departure city ID
        destination_id (str): Arrival city ID
    """

    origin_id: str
    destination_id: str


@dataclass
class CityRoute(BoundedItem, DerivedItem, CustomersModelHolder, AirlineLocation):
    """Pair of cities with unspecified direction.

    Args:
        first_id (str): First city ID of the pair in alphabetical order
        second_id (str): Second city ID of the pair in alphabetical order
    """

    first_id: str
    second_id: str


@dataclass
class Airport(BoundedItem, CoreItem):
    """Airport instance.

    Args:
        name (str): Full airport name
        city_id (str): City ID of the airport
        iata (str): IATA code of the airport
        icao (str): ICAO code of the airport
        altitude (int): Location altitude in MAMSL
        latitude (int): Location latitude
        longitude (int): Location longitude
    """

    name: str
    city_id: str
    iata: str
    icao: str = "undefined"
    altitude: int = 0
    latitude: int = 0
    longitude: int = 0

    def __post_init__(self):
        self.altitude = int(self.altitude)
        self.latitude = int(self.latitude)
        self.longitude = int(self.longitude)


@dataclass
class Sector(BoundedItem, DerivedItem, AirlineLocation):
    """Pair of airports with specified direction.

    Args:
        origin_id (str): Departure airport ID
        destination_id (str): Arrival airport ID
    """

    origin_id: str
    destination_id: str
    citysector_id: str


@dataclass
class Route(BoundedItem, DerivedItem, AirlineLocation):
    """Pair of airports with unspecified direction.

    Args:
        first_id (str): First airport ID of the pair in alphabetical order
        second_id (str): Second airport ID of the pair in alphabetical order
    """

    first_id: str
    second_id: str
    cityroute_id: str


@dataclass
class Schedule(UnboundedItem, CoreItem):
    """Schedule instance. Template generator for flights in an airline with common aircraft, sector and days of week.

    Args:
        airline_id (str): Airline ID
        aircraft_id (str): Aircraft ID
        origin_id (str): Departure airport ID
        destination_id (str): Arrival airport ID
        days_of_week (List[DayOfWeek]): Days of week
        departure_time (timedelta): Departure time upto minutes (hour-minutes)
        duration (timedelta): Flight duration upto minutes (hour-minutes)
        flight_number (int): Flight number identifier
        sell_before_days (int): Flights are put on sale these days before departure day
        from_date (datetime): Start date of the range of flight departures, upto days (year-month-day)
        to_date (datetime): End date of the range of flight departures, up to days (year-month-day)
        sector_id (str): Sector ID
        citysector_id (str): CitySector ID
        route_id (str): Route ID
        cityroute_id (str): CityRoute ID
        from_date_load (datetime): Start date of the selling period, upto days (year-month-day)
        to_date_load (datetime): End date of the selling period, upto days (year-month-day)
        from_date_departure (datetime): Date of the first departure, upto days (year-month-day)
        to_date_departure (datetime): Date of the last departure, up to days (year-month-day)

    """

    airline_id: str
    aircraft_id: str
    origin_id: str
    destination_id: str
    days_of_week: List[DayOfWeek]
    departure_time: timedelta
    duration: int
    flight_number: str
    sell_before_days: int
    from_date: str
    to_date: str
    sector_id: str
    citysector_id: str
    route_id: str
    cityroute_id: str
    from_date_load: datetime
    to_date_load: datetime
    from_date_departure: datetime
    to_date_departure: datetime

    def __post_init__(self):

        self.sell_before_days = int(self.sell_before_days)

        self.days_of_week = [
            DayOfWeek.int_to_enum_value(int(d))
            for d in self.days_of_week.replace("-", "")
        ]
        self.departure_time = self.departure_time

        self.from_date = datetime.strptime(self.from_date, DateFormat)
        self.to_date = datetime.strptime(self.to_date, DateFormat)
        self.from_date_load = datetime.strptime(self.from_date_load, DateFormat)
        self.to_date_load = datetime.strptime(self.to_date_load, DateFormat)
        self.from_date_departure = datetime.strptime(
            self.from_date_departure, DateFormat
        )
        self.to_date_departure = datetime.strptime(self.to_date_departure, DateFormat)

        self.departure_time = timedelta(minutes=int(self.departure_time))
        self.duration = timedelta(minutes=int(self.duration))


@dataclass
class Flight(UnboundedItem, DerivedItem, PricingModelHolder):
    """Flight instance.

    Args:
        seat_capacity (int): Seat capacity
        schedule_id (str): Schedule ID
        sector_id (str): Sector ID
        citysector_id (str): CitySector ID
        airline_id (str): Airline ID
        aircraft_id (str): Aircraft ID
        flight_number (str): Flight number identifier
        onsale_date_time (datetime): Flight is put on sale this date, upto days (year-month-day)
        departure_date_time (datetime): Departure date and time, upto minutes (year-month-day-hour-minute)
        fares (List[str]): List of fare identifiers
        lowest_pps (int): Price per seat of the lowest fare in cents
        highest_pps (int): Price per seat of the highest fare in cents
    """

    seat_capacity: int
    schedule_id: str
    sector_id: str
    citysector_id: str
    airline_id: str
    aircraft_id: str
    flight_number: str
    onsale_date_time: datetime
    departure_date_time: datetime
    fares: List[str]
    lowest_pps: int
    highest_pps: int

    def __post_init__(self):
        self.seat_capacity = int(self.seat_capacity)
        self.fares = self.fares.split("-")
        self.lowest_pps = self.lowest_pps
        self.highest_pps = self.highest_pps
        self.onsale_date_time = datetime.strptime(self.onsale_date_time, DateFormat)
        self.departure_date_time = datetime.strptime(
            self.departure_date_time, DateTimeMinFormat
        )


AnyAirlineLocationType = Union[Sector, CitySector, Route, CityRoute]
AnyBoundedType = Union[
    PModel,
    Airline,
    Aircraft,
    Country,
    City,
    CitySector,
    CityRoute,
    Airport,
    Sector,
    Route,
]
AnyUnboundedType = Union[Schedule, Flight]

AllBoundedItems = [
    obj
    for name, obj in inspect.getmembers(importlib.import_module(__name__))
    if inspect.isclass(obj) and obj != BoundedItem and issubclass(obj, BoundedItem)
]

AllUnboundedItems = [
    obj
    for name, obj in inspect.getmembers(importlib.import_module(__name__))
    if inspect.isclass(obj) and obj != UnboundedItem and issubclass(obj, UnboundedItem)
]

AllCoreItems = [
    obj
    for name, obj in inspect.getmembers(importlib.import_module(__name__))
    if inspect.isclass(obj) and obj != CoreItem and issubclass(obj, CoreItem)
]

AllDerivedItems = [
    obj
    for name, obj in inspect.getmembers(importlib.import_module(__name__))
    if inspect.isclass(obj) and obj != DerivedItem and issubclass(obj, DerivedItem)
]

AirlineLocationItems = [
    obj
    for name, obj in inspect.getmembers(importlib.import_module(__name__))
    if inspect.isclass(obj)
    and obj != AirlineLocation
    and issubclass(obj, AirlineLocation)
]



"""Interface for core data uploading. """

import os
from typing import Optional
from rmlab._api.upload import APIUploadInternal
from rmlab.data.items import (
    Aircraft,
    Airline,
    Airport,
    City,
    Country,
    Schedule,
)


class APIUploadCore(APIUploadInternal):
    """Exposes functions for uploading local data to the server."""

    async def upload_aircrafts(self, scen_id: int, data_fn: str) -> None:
        """Upload a set of aircrafts defined in a file.


        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename `.csv` or `.json` defining the aircrafts

        Where `data_fn` can refer to:

        * a **CSV file**, for instance:
        ```csv
        Model,Seat capacity
        Airbus A320-a,174
        Airbus A320-b,180
        ```

        * a **JSON file**, for instance:
        ```json
        [
            {"model": "Airbus A320-a", "seat_capacity": 174},
            {"model": "Airbus A320-b", "seat_capacity": 180},
        ]
        ```

        Raises:
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
        """

        await self._upload_bounded_items(
            scen_id=scen_id, category=Aircraft, data_fn=data_fn
        )

    async def upload_airlines(self, scen_id: int, data_fn: str) -> None:
        """Upload a set of airlines defined in a file.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename `.csv` or `.json` defining the airlines

        Where `data_fn` can refer to:

        * a **CSV file**, for instance:
        ```csv
        Name,Type
        MyCarrier,low-cost
        AnotherCarrier,low-cost
        MyLegacyCarrier,legacy
        ```

        * a **JSON file**, for instance:
        ```json
        [
            {"name": "MyCarrier", "type": "low-cost"},
            {"name": "AnotherCarrier", "type": "low-cost"},
            {"name": "MyLegacyCarrier", "type": "legacy"},
        ]
        ```

        Raises:
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
        """

        await self._upload_bounded_items(
            scen_id=scen_id, category=Airline, data_fn=data_fn
        )

    async def upload_airports(self, scen_id: int, data_fn: str) -> None:
        """Upload a set of airports defined in a file.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename `.csv` or `.json` defining the airports

        Where `data_fn` can refer to:

        * a **CSV file**, for instance:
        ```csv
        Name,City,Altitude,Latitude,Longitude,ICAO,IATA
        Geneva,Geneva,431,46.238,6.109,LSGG,GVA
        Madrid,Madrid,619,40.294,-3.724,LEMD,MAD
        London Gatwick,London,60,51.148,-0.19,EGKK,LGW
        London Heathrow,London,25,51.477,-0.461,EGLL,LHR
        ```

        * a **JSON file**, for instance:
        ```json
        [
            {"name": "Geneva", "city": "Geneva", "altitude": 431, "latitude": 46.238, "longitude": 6.109, "icao": "LSGG", "iata": "GVA"},
            {"name": "Madrid", "city": "Madrid", "altitude": 619, "latitude": 40.294, "longitude": -3.724, "icao": "LEMD", "iata": "MAD"},
            {"name": "London Gatwick", "city": "London", "altitude": 60, "latitude": 51.148, "longitude": -0.19, "icao": "EGKK", "iata": "LGW"},
            {"name": "London Heathrow", "city": "London", "altitude": 25, "latitude": 51.477, "longitude": -0.461, "icao": "EGLL", "iata": "LHR"},
        ]
        ```

        * **NOTE**: Airport *cities* must reference previously uploaded ``City`` items with the same `City.name`.

        Raises:
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
            rmlab_errors.ClientError: If any of the referenced cities does not exist in server
        """

        await self._upload_bounded_items(
            scen_id=scen_id, category=Airport, data_fn=data_fn
        )

    async def upload_cities(self, scen_id: int, data_fn: str) -> None:
        """Upload a set of cities defined in a file.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename `.csv` or `.json` defining the cities

        Where ``data_fn`` can refer to:

        * a **CSV file**, for instance:
        ```csv
        Name,Country
        Geneva,Swizterland
        Madrid,Spain
        London,United Kingdom
        ```

        * a **JSON file**, for instance:
        ```json
        [
            {"name": "Geneva", "country": "Switzerland"},
            {"name": "Madrid", "country": "Spain"},
            {"name": "London", "country": "United Kingdom"},
        ]
        ```

        * **NOTE**: City *countries* must reference previously uploaded ``Country`` items with the same `Country.name`.

        Raises:
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
            rmlab_errors.ClientError: If any of the referenced countries does not exist in server
        """

        await self._upload_bounded_items(
            scen_id=scen_id, category=City, data_fn=data_fn
        )

    async def upload_countries(self, scen_id: int, data_fn: str) -> None:
        """Upload a set of countries defined in a file.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename `.csv` or `.json` defining the countries

        Where `data_fn` can refer to:

        * a **CSV file**, for instance:
        ```csv
        Name,Currency
        Swizterland,chf
        Spain,eur
        United Kingdom,gbp
        ```

        * a **JSON file**, for instance:
        ```json
        [
            {"name": "Switzerland", "currency": "chf"},
            {"name": "Spain", "currency": "eur"},
            {"name": "United Kingdom", "currency": "gbp"},
        ]
        ```

        Raises:
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
        """

        await self._upload_bounded_items(
            scen_id=scen_id, category=Country, data_fn=data_fn
        )

    async def upload_schedules(
        self,
        scen_id: int,
        data_fn: str,
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> None:
        """Upload a set of flights schedules defined in a file.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename `.csv` or `.json` defining the schedules
            citysector_id (Optional[str], optional): A citysector ID if all schedules belong to the same citysector. Defaults to None.
            sector_id (Optional[str], optional): A sector ID if all schedules belong to the same sector. Defaults to None.

        Arguments `citysector_id` and `sector_id` are always optional. If provided, concurrent uploads of schedules
            files belonging to different citysectors/sectors are allowed.

        Where `data_fn` can refer to:

        * a **CSV file**, for instance:
        ```csv
        Airline,Aircraft,Origin,Destination,Days of Week,Departure time,Duration,Flight number,Sell before days,From Date,To Date
        MyCarrier,Airbus A320-b,MAD,GVA,-2-4-6-,T08:00,T02:00,2277,15,2022-04-01,2022-05-31
        MyCarrier,Airbus A320-a,MAD,GVA,1-3-5-7,T16:00,T02:00,2278,15,2022-04-01,2022-05-31
        MyCarrier,Airbus A320-b,GVA,MAD,-2-4-6-,T10:30,T01:30,2377,15,2022-04-01,2022-05-31
        MyCarrier,Airbus A320-a,GVA,MAD,1-3-5-7,T18:30,T01:30,2378,15,2022-04-01,2022-05-31
        ```

        * a **JSON file**, for instance:
        ```json
        [
            {"airline": "MyCarrier", "aircraft": "Airbus A320-b", "origin": "MAD", "destination": "GVA", "days_of_week": "-2-4-6-", "departure_time": "T08:00", "duration": "T02:00", "flight_number": "2277", "sell_before_days": 45, "from_date": "2022-04-01", "to_date": "2022-05-31"},
            {"airline": "MyCarrier", "aircraft": "Airbus A320-b", "origin": "MAD", "destination": "GVA", "days_of_week": "1-3-5-7", "departure_time": "T16:00", "duration": "T02:00", "flight_number": "2278", "sell_before_days": 45, "from_date": "2022-04-01", "to_date": "2022-05-31"},
            {"airline": "MyCarrier", "aircraft": "Airbus A320-b", "origin": "GVA", "destination": "MAD", "days_of_week": "-2-4-6-", "departure_time": "T10:30", "duration": "T01:30", "flight_number": "2377", "sell_before_days": 45, "from_date": "2022-04-01", "to_date": "2022-05-31"},
            {"airline": "MyCarrier", "aircraft": "Airbus A320-b", "origin": "GVA", "destination": "MAD", "days_of_week": "1-3-5-7", "departure_time": "T10:30", "duration": "T01:30", "flight_number": "2378", "sell_before_days": 45, "from_date": "2022-04-01", "to_date": "2022-05-31"},
        ]
        ```

        * **NOTE**: Schedule *airlines* must reference previously uploaded ``Airline`` items with the same `Airline.name`.
        * **NOTE**: Schedule *aircrafts* must reference previously uploaded ``Aircraft`` items with the same `Aircraft.model`.
        * **NOTE**: Schedule *origins* must reference previously uploaded ``Airport`` items with the same `Airport.iata`.
        * **NOTE**: Schedule *destinations* must reference previously uploaded ``Airport`` items with the same `Airport.iata`.
        * **NOTE**: No two schedules with overlapping date ranges (*from_date*, *to_date*), and overlapping days of week, with the same *flight_number* are allowed.


        Raises:
            ValueError: If file extension is invalid
            FileNotFoundError: If file does not exist
            rmlab_errors.ClientError: If any of the previous Notes are not satisfied
        """

        await self._upload_unbounded_items(
            scen_id,
            Schedule,
            data_fn,
            citysector_id=citysector_id,
            sector_id=sector_id,
        )

    async def upload_batch_core(
        self,
        scen_id: int,
        *,
        aircraft_items_fn: Optional[str] = None,
        airline_items_fn: Optional[str] = None,
        airport_items_fn: Optional[str] = None,
        city_items_fn: Optional[str] = None,
        country_items_fn: Optional[str] = None,
        schedule_items_fn: Optional[str] = None,
    ):
        """Upload a set of items defined in files to server.

        Args:
            scen_id (int): Scenario ID
            aircraft_items_fn (Optional[str], optional): File name with aircraft items. Defaults to None.
            airline_items_fn (Optional[str], optional): File name with airline items. Defaults to None.
            airport_items_fn (Optional[str], optional): File name with airport items. Defaults to None.
            city_items_fn (Optional[str], optional): File name with city items. Defaults to None.
            country_items_fn (Optional[str], optional): File name with country items. Defaults to None.
            schedule_items_fn (Optional[str], optional): File name with flights schedule items. Defaults to None.

        Raises:
            FileNotFoundError: If any of the files does not exist
        """

        not_existing_fns = [
            fn
            for fn in [
                airport_items_fn,
                airline_items_fn,
                airport_items_fn,
                city_items_fn,
                country_items_fn,
                schedule_items_fn,
            ]
            if fn and not os.path.exists(fn)
        ]

        if len(not_existing_fns) > 0:
            raise FileNotFoundError(not_existing_fns)

        if airline_items_fn:
            await self._upload_bounded_items(scen_id, Airline, data_fn=airline_items_fn)

        if aircraft_items_fn:
            await self._upload_bounded_items(
                scen_id, Aircraft, data_fn=aircraft_items_fn
            )

        if country_items_fn:
            await self._upload_bounded_items(scen_id, Country, data_fn=country_items_fn)

        if city_items_fn:
            await self._upload_bounded_items(scen_id, City, data_fn=city_items_fn)

        if airport_items_fn:
            await self._upload_bounded_items(scen_id, Airport, data_fn=airport_items_fn)

        if schedule_items_fn:
            await self._upload_unbounded_items(
                scen_id, Schedule, data_fn=schedule_items_fn
            )



from rmlab.api import API

from rmlab.data.items import (
    Aircraft,
    Airline,
    Airport,
    City,
    CityRoute,
    CitySector,
    Country,
    Route,
    Sector,
    Schedule,
    Flight,
)

from rmlab.data.flight import (
    FlightDataBooks,
    FlightDataEvents,
    FlightDataForecastedBooks,
    FlightDataPricePerSeatSettings,
    FlightDataThresholdSettings,
)

from rmlab.data.scenario import (
    ScenarioDates,
    ItemsCount,
    SchedulesCount,
    FlightsCount,
)

__all__ = [
    "API",
    "Aircraft",
    "Airline",
    "Airport",
    "City",
    "CityRoute",
    "CitySector",
    "Country",
    "Route",
    "Sector",
    "Schedule",
    "Flight",
    "FlightDataBooks",
    "FlightDataEvents",
    "FlightDataForecastedBooks",
    "FlightDataPricePerSeatSettings",
    "FlightDataThresholdSettings",
    "ScenarioDates",
    "ItemsCount",
    "SchedulesCount",
    "FlightsCount",
]



from typing import List
from rmlab._api.observability import APIObservabilityInternal


class APIObservability(APIObservabilityInternal):
    """Interface to fetch monitoring data from server."""

    async def fetch_logs(self, context: str, count: int) -> List[str]:
        """Fetch last 'count' monitor logs from the server.

        Args:
            context (str): Logging context, either 'info', 'warning', 'error' or 'critical'
            count (int): Maximum number of log messages. Must be positive.
        
        Raises:
            ValueError: If 'context' or 'count' don't satisfy requirements.

        Returns:
            A list of log messages.
        """

        if context.lower() not in ["info", "warning", "error", "critical"]:
            raise ValueError(f"Expected log context either 'info', 'warning', 'error' or 'critical'")
        
        if not isinstance(count, int) or (isinstance(count, int) and count <= 0):
            raise ValueError(f"Expected 'count' to be a positive integer")

        return await self._fetch_logs(context, count)



from typing import List
from rmlab._api.base import APIBaseInternal

class APIObservabilityInternal(APIBaseInternal):
    """Interface to fetch monitoring data from server."""

    async def _fetch_logs(self, context: str, count: int) -> List[str]:

        return await self._submit_call("api-monitor-activity-user", context=context, count=count)


from typing import List, Optional, Union
from rmlab._api.upload import APIUploadInternal
from rmlab.data.items import (
    Flight,
)
from rmlab._util import run_async_tasks
from rmlab.data.flight import (
    FlightDataBooks,
    FlightDataPricePerSeatSettings,
    FlightDataThresholdSettings,
)


class APIUploadFlightData(APIUploadInternal):
    """Interface to upload flight data to server"""

    async def upload_flights(
        self,
        scen_id: int,
        data_fn: str,
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> None:
        """Upload a set of flights defined in a file.

        Args:
            scen_id (int): Scenario ID
            data_fn (str): Filename (.csv or .json) defining the flights
            citysector_id (Optional[str], optional): A citysector ID if all flights belong to the same citysector. Defaults to None.
            sector_id (Optional[str], optional): A sector ID if all flights belong to the same sector. Defaults to None.

        Arguments `citysector_id` and `sector_id` are always optional. If any of them provided, concurrent uploads of flights
            files belonging to different citysectors/sectors are allowed.

        Where `data_fn` can refer to:
        * a **CSV file**, for instance:
        ```csv
        Airline,Aircraft,Origin,Destination,Flight number,On sale date,Departure date,Departure time,Duration
        MyCarrier,Airbus A320-b,MAD,GVA,2277,2022-04-01,2022-05-15,T16:30,T02:00
        MyCarrier,Airbus A320-b,MAD,GVA,2277,2022-04-01,2022-05-17,T20:30,T02:00
        ```

        * a **JSON file**, for instance:
        ```json
        [
            {"airline": "MyCarrier", "aircraft": "Airbus A320-b", "origin": "MAD", "destination": "GVA", "flight_number": "2277", "on_sale_date": "2022-04-01", "departure_date": "2022-05-15", "departure_time": "T16:30", "duration": "T02:00", },
            {"airline": "MyCarrier", "aircraft": "Airbus A320-b", "origin": "MAD", "destination": "GVA", "flight_number": "2277", "on_sale_date": "2022-04-01", "departure_date": "2022-05-17", "departure_time": "T20:30", "duration": "T02:00", },
        ]
        ```
        """

        await self._upload_unbounded_items(
            scen_id, Flight, data_fn, citysector_id=citysector_id, sector_id=sector_id
        )

    async def upload_flights_data(
        self,
        scen_id: int,
        flights_data: List[
            Union[
                FlightDataBooks,
                FlightDataThresholdSettings,
                FlightDataPricePerSeatSettings,
            ]
        ],
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ) -> None:
        """Upload flight data of multiple flights belonging to a citysector or sector to server.

        Any previous existing data is overwritten.
        At least `citysector_id` and/or `sector_id` associated to the flights must be defined.

        Args:
            scen_id (int): Scenario ID
            flights_data (List[Union[FlightDataBooks, FlightDataThresholdSettings, FlightDataPricePerSeatSettings]]): List of flights data to upload.
            citysector_id (Optional[str], optional): Target citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Target sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined
            ValueError: If one of flight data list elements has incorrect type.
            MultipleError: If several of flight data list elements have incorrect types.
        """

        if citysector_id is None and sector_id is None:
            raise ValueError(
                f"Require definition of either `citysector_id` or `sector_id`"
            )

        endpoint_limit = self._api_endpoints_limits.data_flight_post

        if len(flights_data) > endpoint_limit:

            # Sequentially upload data in chunks to keep the request payloads and server load controlled

            for chunk_start in range(0, len(flights_data), endpoint_limit):

                chunk_end = chunk_start + endpoint_limit
                if chunk_end > len(flights_data):
                    chunk_end = -1

                await run_async_tasks(
                    [
                        self._upload_flight_data(
                            scen_id,
                            flight_data,
                            citysector_id=citysector_id,
                            sector_id=sector_id,
                        )
                        for flight_data in flights_data[chunk_start:chunk_end]
                    ],
                    return_results=False,
                )

        else:

            await run_async_tasks(
                [
                    self._upload_flight_data(
                        scen_id,
                        flight_data,
                        citysector_id=citysector_id,
                        sector_id=sector_id,
                    )
                    for flight_data in flights_data
                ],
                return_results=False,
            )



"""Interface for fetching data related to parametric models and filters."""

from typing import List
from rmlab.data.items import PModel
from rmlab.data.parametric.filter import PFilter
from rmlab._api.fetch import APIFetchInternal


class APIFetchParametric(APIFetchInternal):
    """Exposes functions for fetching flight data from the server."""

    async def fetch_parametric_filters(self, scen_id: int) -> List[PFilter]:
        """Fetch a list of all parametric filters of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of parametric filters
        """

        return await self._fetch_bounded_items(scen_id, PFilter)

    async def fetch_parametric_models(self, scen_id: int) -> List[PModel]:
        """Fetch a list of all parametric models of scenario from server.

        Args:
            scen_id (int): Scenario ID

        Returns:
            List of parametric models
        """

        return await self._fetch_bounded_items(scen_id, PModel)



"""Interface to run simulations and set simulation checkpoint dates."""

import asyncio
from re import A
from typing import Mapping

from tqdm import tqdm
from rmlab._api.fetch import APIFetchInternal


class APISimulationInternal(APIFetchInternal):
    """Exposes functions for simulation."""

    async def _periodic_trigger_progress(
        self, scen_id: int, stop_event: asyncio.Event
    ) -> None:
        """Emits progress bars in the console to track the progress of the simulation.

        Args:
            scen_id (int): Scenario ID in which the simulation is running.
            stop_event (asyncio.Event): Event to notify the stop of the simulation.
        """

        pbars: Mapping[str, tqdm] = dict()

        try:

            (
                dates,
                bounded_count,
                schedules_count,
                flights_count,
            ) = await self._fetch_info(scen_id)

            pbars = {
                "local": tqdm(
                    desc=f"Local progress (SC{scen_id})",
                    unit="day",
                    total=(dates.next - dates.current).days,
                    initial=0,
                ),
                "global": tqdm(
                    desc=f"Global progress (SC{scen_id})",
                    unit="day",
                    total=(dates.last_flight_departure - dates.first_flight_load).days,
                    initial=(dates.current - dates.first_flight_load).days,
                ),
            }

            while not stop_event.is_set():

                await asyncio.sleep(2)

                (
                    dates,
                    bounded_count,
                    schedules_count,
                    flights_count,
                ) = await self._fetch_info(scen_id)

                # TODO: Make these prints compatible with progress bars:
                # print(f"Flights (past/live/pending): {flights_count.past}/{flights_count.live}/{flights_count.pending}              ", end='\r')
                # print(f"Schedules (past/live/pending): {schedules_count.past}/{schedules_count.live}/{schedules_count.pending}              ", end='\r')
                pbars["local"].n = (
                    pbars["local"].total - (dates.next - dates.current).days
                )
                pbars["global"].n = (
                    pbars["global"].total
                    - (dates.last_flight_departure - dates.current).days
                )

                for pb in pbars.values():
                    pb.update()

        except Exception:

            raise

        finally:

            for pb in pbars.values():
                pb.close()



"""Interface to trigger optimization passes."""

from datetime import datetime
from typing import Optional
from rmlab._api.base import APIBaseInternal
from rmlab._data.types import DateTimeMinFormat


class APIOptimization(APIBaseInternal):
    """Exposes functions to run and schedule optimization runs on a set of flights."""

    async def trigger_optimization_pass(
        self,
        scen_id: int,
        airline_id: str,
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ):
        """Triggers an optimization pass on all flights of an airline belonging to a citysector or sector.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID
            citysector_id (Optional[str], optional): Citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined
        """

        if citysector_id is None and sector_id is None:
            raise ValueError(
                f"At least one of `citysector_id`, `sector_id` must be defined"
            )

        await self._submit_call(
            "api-operation-optimize-trigger",
            scen_id=scen_id,
            airline_id=airline_id,
            citysector_id=citysector_id,
            sector_id=sector_id,
        )

    async def schedule_optimization_pass(
        self,
        scen_id: int,
        airline_id: str,
        date_time: datetime,
        *,
        citysector_id: Optional[str] = None,
        sector_id: Optional[str] = None,
    ):
        """Schedules an optimization pass on all flights of an airline belonging to a citysector or sector to be run at specific date and time.

        Args:
            scen_id (int): Scenario ID
            airline_id (str): Airline ID
            date_time (datetime): Date and time at which the optimization pass is triggered.
            citysector_id (Optional[str], optional): Citysector ID. Defaults to None.
            sector_id (Optional[str], optional): Sector ID. Defaults to None.

        Raises:
            ValueError: If none of `citysector_id`, `sector_id` is defined
        """

        if citysector_id is None and sector_id is None:
            raise ValueError(
                f"At least one of `citysector_id`, `sector_id` must be defined"
            )

        await self._submit_call(
            "api-operation-optimize-schedule",
            scen_id=scen_id,
            date_time=datetime.strftime(date_time, DateTimeMinFormat),
            airline_id=airline_id,
            citysector_id=citysector_id,
            sector_id=sector_id,
        )



"""Script exposing enumeration types and values to add a type/value-safe 
layer for correctness."""

from enum import Enum


class LowerStringEnumValues(Enum):
    @classmethod
    def str_to_enum_value(cls, v: str):
        return cls.__dict__["_value2member_map_"][v.lower()]


class FileExtensions(LowerStringEnumValues):
    """All possible extensions of uploadable files"""

    JSON = ".json"
    CSV = ".csv"


class DataRemoveKind(LowerStringEnumValues):
    COMPLETE = "full"
    HISTORIC = "restart"


class CurrencyKind(LowerStringEnumValues):
    """All recognized currencies"""

    CROATIAN_KUNA = "hrk"
    CZECH_KORUNA = "czk"
    DANISH_KRONE = "dkk"
    EURO = "eur"
    HUNGARIAN_FORINT = "hug"
    MOLDOVAN_KEY = "mdl"
    NORWEGIAN_KRONE = "nok"
    POLISH_ZLOTY = "pln"
    POUND_STERLING = "gbp"
    SWEDISH_KRONA = "sek"
    SWISS_FRANC = "chf"
    UKRAINIAN_HRYVNIA = "uah"
    US_DOLLAR = "usd"


class ParametricModelKind(LowerStringEnumValues):
    """All possible enumerations for types of parametric models"""

    CUSTOMERS = "customers"
    PRICING = "pricing"


class CustomersModelKind(LowerStringEnumValues):
    """All possible enumerations for types of customers models"""

    REQUEST = "request"
    CHOICE = "choice"


class PricingModelKind(LowerStringEnumValues):
    """All possible enumerations for types of pricing models"""

    RANGE = "range"
    BEHAVIOR = "behavior"
    OPTIMIZER = "optimizer"


class DayOfWeek(Enum):
    """Days of week enumeration"""

    MONDAY = 1
    TUESTDAY = 2
    WEDNESDAY = 3
    THURSDAY = 4
    FRIDAY = 5
    SATURDAY = 6
    SUNDAY = 7

    @classmethod
    def int_to_enum_value(cls, v: int):
        return cls.__dict__["_value2member_map_"][v]


class ScenarioDayStatus(LowerStringEnumValues):
    """All possible enumerations of daily status"""

    EMPTY = "empty"
    STARTED = "started"
    ENDED = "ended"
    IN_PROGRESS = "inprogress"
    READY = "ready"
    PAUSED = "paused"
    SIMULATED = "simulated"
    OPTIMIZED = "optimized"


class ScenarioState(LowerStringEnumValues):
    """All possible enumerations of scenario states"""

    EMPTY = "empty"
    IN_PROGRESS = "inprogress"
    PAUSED = "paused"
    FINISHED = "finished"


class FlightDataKind(LowerStringEnumValues):
    """All possible enumerations for types of flight data"""

    ACTUAL_BOOKS = "actual-books"
    EXPECTED_BOOKS = "expected-books"
    DYNAMIC_BOOKS = "dynamic-books"
    THRESHOLDS_SETTINGS = "thresholds-settings"
    PRICE_PER_SEAT_SETTINGS = "pps-settings"
    EVENTS = "events"


class FlightEvent(LowerStringEnumValues):
    """All possible enumerations for flight events"""

    FLIGHT_LOADED = "flight-loaded"
    FLIGHT_DEPARTED = "flight-departed"
    FLIGHT_SOLD_OUT = "flight-sold-out"
    SCHEDULED_OPT_PASS = "scheduled-opt-pass"
    DONE_OPT_PASS = "done-opt-pass"
    COMMIT_OPT_PASS = "commit-opt-pass"
    LOADED_INITIAL_THRESHOLDS = "loaded-initial-thresholds"
    OPERATOR_EDIT_THRESHOLDS = "operator-edit-thresholds"
    OPERATOR_DYNAMIC_TO_EXPECTED = "operator-dynamic-to-expected"


class OptimizationSelectorFilterKind(LowerStringEnumValues):
    """All possible enumerations for optimization selector kinds."""

    DAY_OF_WEEK = "day-of-week"
    HOUR_SLOT = "hour-slot"
    SECTOR = "sector"
    SCHEDULE = "schedule"
    FLIGHT_NUMBER = "flight-number"
    SEAT_CAPACITY = "seat-capacity"


class TimeUnit(Enum):
    """Enumerations for time units in optimization selector."""

    DAY = "day"
    MONTH = "month"
    YEAR = "year"

    @classmethod
    def str_to_enum_value(cls, v: str):
        # Make it case-insensitive
        v = v.lower()
        if v[-1] == "s":  # Make it singular
            v = v[0:-1]
        return cls.__dict__["_value2member_map_"][v]


class OptimizationForecasterKind(LowerStringEnumValues):
    """All implemented forecaster kinds."""

    Q_FORECAST = "q-forecast"
    BAYES = "bayes"


class OptimizationAggregatorKind(LowerStringEnumValues):
    """All implemented ways to aggregate historic data."""

    UNIFORM = "uniform"
    EXPONENTIAL = "exponential"


class OptimizationMaximizerKind(LowerStringEnumValues):
    """All implemented revenue maximization algorithms."""

    DP_QSD = "qsd"


class OptimizationEffectsKind(LowerStringEnumValues):
    """All implemented post-optimization effects."""

    NONE = "none"
    COMMIT_THRESHOLDS = "commit"


class ParametricFilterClassKind(LowerStringEnumValues):
    """All implemented filter class kinds."""

    IN_PERIOD = "in_period"
    FLIGHT_NUMBER = "flight_number"
    AIRLINE = "airline"
    SECTOR = "sector"
    ROUTE = "route"
    CITYSECTOR = "citysector"
    CITYROUTE = "cityroute"
    NETWORK = "network"



"""A pricing range defines the amount of fare prices, the min-max range of prices, and the fare identifications."""

from dataclasses import dataclass
import json
from typing import List, Union
from rmlab._data.conversions import parse_data_from_json
from rmlab._data.enums import FileExtensions, PricingModelKind
from rmlab.data.parametric._pricing_base import PricingModel


@dataclass
class RangeModel(PricingModel):
    """Dataclass for pricing range parameters.Parameters specifying .

    Args:
        granularity (int): Number of fares
        min_fare (int): Price per seat of the lowest fare
        max_fare (int): Price per seat of the highest fare
        tags (List[str]): List of single-character identifiers for all fares. Defaults

    Example
    ```py
    range = RangeModel(granularity=4, min_fare=40, max_fare=200, tags=["D", "C", "B", "A"])
    ```

    """

    granularity: int
    min_fare: float
    max_fare: float
    tags: List[str]

    def __post_init__(self):

        if len(self.tags) != self.granularity:
            raise ValueError(
                f"Number of tag identifiers `{len(self.tags)}` must match granularity `{self.granularity}`"
            )


def make_pricing_range_from_json(filename_or_dict: Union[str, dict]) -> RangeModel:
    """Make a pricing range instance from a json representation (from file or dict).

    Args:
        filename_or_dict (Union[str, dict]): JSON filename or dictionary in json format

    Examples from dict:
    ```py
    dict_range = {"granularity": 3, "min_fare": 40, "max_fare": 200}
    my_pricing_range = make_from_json(dict_range)
    # Tags will be ["Q", "P", "O"] from lowest to highest price per seat
    ```

    ```py
    dict_range = {"granularity": 3, "min_fare": 40, "max_fare": 200, tags=["C", "B", "A"]}
    my_pricing_range = make_from_json(dict_range)
    ```

    ```py
    dict_range = {"granularity": 3, "min_fare": 40, "max_fare": 200, lowest_tag="Q"}
    my_pricing_range = make_from_json(dict_range)
    # Tags will be ["Q", "P", "O"] from lowest to highest price per seat
    ```

    ```py
    dict_range = {"granularity": 10, "min_fare": 40, "max_fare": 200, highest_tag="A"}
    my_pricing_range = make_from_json(dict_range)
    # Tags will be ["C", "B", "A"] from lowest to highest price per seat
    ```

    Example from file:
    `my_pricing_range.json`
    ```json
    {
      "granularity" : 10,
      "min_fare" : 40,
      "max_fare" : 500,
      "lowest_tag" : "Q"
    }
    ```
    ```py
    my_pricing_range = make_pricing_range_from_json("my_pricing_range.json")
    ```
    """

    model_id, filename, md5hash, content = parse_data_from_json(filename_or_dict)

    if not isinstance(content, dict):
        raise TypeError(f"Expected dict format in {filename_or_dict}")

    content: dict

    tags = None
    if "tags" in content:
        tags = content["tags"]
    elif "lowest_tag" in content:
        lt = content["lowest_tag"]
        if len(lt) != 1:
            raise ValueError(f"`lowest_tag` must be a single character")
        tags = [chr(ord(lt) - t) for t in range(content["granularity"])]
    elif "highest_tag" in content:
        gt = content["highest_tag"]
        if len(gt) != 1:
            raise ValueError(f"`highest_tag` must be a single character")
        tags = [chr(ord(gt) + t) for t in range(content["granularity"])]
    else:
        tags = [chr(ord("Q") - t) for t in range(content["granularity"])]
        # Defaults to ['Q', 'P', 'O', ... if none provided, with 'Q' identifying the lowest fare]

    return RangeModel(
        id=model_id,
        filename=filename,
        cnt=json.dumps(content),
        hash=md5hash,
        kind=PricingModelKind.RANGE,
        extension=FileExtensions.JSON,
        granularity=int(content["granularity"]),
        min_fare=float(content["min_fare"]),
        max_fare=float(content["max_fare"]),
        tags=list(tags),
    )



"""A choice model emulates customers selection criteria while choosing flight booking offers.
From a functional perspective, a choice model maps 
a list of booking offers for a given flight to a selected booking offer, 
either deterministically or by sampling from some statistical distribution. 
The selected offer may be none (ie, all offers are rejected), with some probability depending on the type of the model.

Args:
  class_name (str): Identifier of a customers class (eg `business`, `leisure`)
"""

from dataclasses import dataclass
import json
from typing import List, Union
from rmlab._data.conversions import parse_data_from_json
from rmlab._data.enums import CustomersModelKind, FileExtensions

from rmlab.data.parametric._customers_base import CustomersModel


@dataclass
class BaseChoiceModel(CustomersModel):
    """Base dataclass for choice models."""


@dataclass
class Random(BaseChoiceModel):
    """Parameterless model. Picks a single booking offer randomly regardless of any parameter.
    The probability of choosing none is zero.

    Args:
      class_name (str): Customers class

    """

    class_name: str


@dataclass
class Cheapest(BaseChoiceModel):
    """Parameterless model. Picks the booking offer with the minimum price.
    Randomly breaking ties if two offers have the same minimum price.
    The probability of choosing none is zero.

    Args:
      class_name (str): Customers class

    """

    class_name: str


# @dataclass
# class MNLSingleReject(BaseChoiceModel):
#   """
#   A multinomial logistic model is used to score the list of booking offers
#   according to its price.
#
#   Requires a beta parameter that quantifies the sensitivity to price, required to define a function
#   that returns a score
#   $$
#     s = \exp(-p/ \\beta)
#   $$
#
#   in [0,1] for a given positive price per seat *p* of each booking offer. The lesser the price,
#   the closer the score is to 1. The greater the *beta*, the less sensitive to price are the customers.
#
#   When there are more than one booking offers, all the scores of the booking offers
#   are normalized to get a discrete distribution over them. Then this distribution
#   is sampled to pick the winner offer.
#
#   If there is a single booking offer, then there is a single score *s < 1*. In this case,
#   there is a non-zero probability of rejecting this single offer *(s - 1)*.
#
#   This model will cause an inelastic demand when there are several flights competing, since the existence of more than 1 booking offer makes the rejection probability 0, so the customers will absorb all the demand.
#   Thus this model is more realistic when there are no competitors.
#
#   Args:
#     class_name (str): Name identifier of the customers class
#     beta (float): $$\\beta$$
#
#   Example:
#   ```py
#   bus_req_model = MNLSingleReject(class_name="business", beta=140)
#   leis_req_model = MNLSingleReject(class_name="leisure", beta=60)
#   ```
#
#   Naming:
#
#   * *MNL*: MultiNomial Logistic
#
#   * *Single reject*: There is a *rejection* only if there is a *single* offer
#   """
#
#   class_name: str
#   beta: float


@dataclass
class Multinomial(BaseChoiceModel):
    """
    A multinomial logistic model is used to score the list of booking offers according to its price.

    This model specifies a price sensitivity parameter $\\beta$ and a rejection probability
    parameter $prob_{reject}$ in [0, 1], so that the probability of rejecting all offers is always fixed to $prob_{reject}$.

    The probability of accepting any booking offer is thus $1 - prob_{reject}$, and the
    acceptance probabilities of the offers given no-rejection are weighted according to a score *s*
    computed from their prices *p*:
    $$
      s = \exp(-p/ \\beta).
    $$

    Note that regarding $\\beta$ sensitivity, its magnitude corresponds to the *unit*
    currency (eg: *euros*, *gbps*, ..., and **not** *euro cents* or *gbp pennies*, ...).

    Args:
      class_name (str): Customers class
      beta (float): $$\\beta$$
      reject_prob (float): $prob_{reject}$

    """

    class_name: str
    beta: float
    reject_prob: float

    def __post_init__(self):

        if self.beta <= 0:
            raise ValueError(f"Beta must be positive, got `{self.beta}`")

        if self.reject_prob < 0 or self.reject_prob >= 1:
            raise ValueError(f"Reject probability must be in [0, 1)")


@dataclass
class MultipleChoiceModels(BaseChoiceModel):
    """An aggregation of request models for different customer classes.

    Example:
    ```py
    bus_choice_model = Multinomial(class_name="business", beta=30, reject_prob=0.5
    lei_choice_model = Cheapest(class_name="leisure", beta=10)
    rnd_choice_model = Random(class_name="careless")

    mul_chouce_model = MultipleChoiceModels(bus_choice_model, lei_choice_model, rnd_choice_model)
    ```

    JSON representation:
    ```json
    {
      "business" : {"type" : "mnl_multiple_reject","beta": 140.0, "reject_prob": 0.5},
      "leisure": {"type": "cheapest"},
      "careless": {"type": "random"}
    }
    ```
    """

    models: List[BaseChoiceModel]


def make_customers_choice_model_from_json(
    filename_or_dict: Union[str, dict]
) -> BaseChoiceModel:
    """Make a request model instance from a json representation (from file or dict).

    Args:
        filename_or_dict (Union[str, dict]): JSON filename or dictionary in json format

    Raises:
        ValueError: If JSON is not a dict
        ValueError: If model type is invalid
        ValueError: If JSON content is empty

    Returns:
        A customers request model instance


    Example from dict:
    ```py
    my_dict = dict()
    my_dict["business"] = {"type": "poisson", "beta": 30.0, "scale": 400, "lambda_seats": 1.5}
    my_dict["leisure"] = {"type": "poisson", "beta": 10.0, "scale": 300, "lambda_seats": 3}

    my_choice_model = make_from_json(my_dict)
    ```

    Example from file:
    `my_choice_model.json`
    ```json
    {
      "business" : {"type" : "mnl_multiple_reject","beta": 140.0, "reject_prob": 0.5},
      "leisure": {"type": "cheapest"},
      "careless": {"type": "random"}
    }
    ```
    ```py
    my_choice_model = make_customers_choice_model_from_json("my_choice_model.json")
    ```
    """

    model_id, filename, md5hash, content = parse_data_from_json(filename_or_dict)

    if not isinstance(content, dict):
        raise ValueError(f"Expected dict format in {filename_or_dict}")

    content: dict

    if len(content) > 0:
        models = list()
        for class_name, fields in content.items():
            if fields["type"] == "random":
                models.append(
                    Random(
                        id=model_id,
                        filename=filename,
                        cnt=json.dumps(content),
                        hash=md5hash,
                        kind=CustomersModelKind.CHOICE,
                        extension=FileExtensions.JSON,
                        class_name=class_name,
                    )
                )
            elif fields["type"] == "cheapest":
                models.append(
                    Cheapest(
                        id=model_id,
                        filename=filename,
                        cnt=json.dumps(content),
                        hash=md5hash,
                        kind=CustomersModelKind.CHOICE,
                        extension=FileExtensions.JSON,
                        class_name=class_name,
                    )
                )
            elif fields["type"] == "mnl_multiple_reject":
                models.append(
                    Multinomial(
                        id=model_id,
                        filename=filename,
                        cnt=json.dumps(content),
                        hash=md5hash,
                        kind=CustomersModelKind.CHOICE,
                        extension=FileExtensions.JSON,
                        class_name=class_name,
                        beta=float(fields["beta"]),
                        reject_prob=float(fields["reject_prob"]),
                    )
                )
            else:
                raise ValueError(f'Unknown request model type {fields["type"]}')

        if len(models) > 1:
            return MultipleChoiceModels(
                id=model_id,
                filename=filename,
                cnt=json.dumps(content),
                hash=md5hash,
                kind=CustomersModelKind.REQUEST,
                extension=FileExtensions.JSON,
                models=models,
            )
        else:
            return models[0]

    else:
        raise ValueError(f"Empty content")



"""
This script provides dataclasses that represent a global information of scenarios.
"""

from dataclasses import dataclass, field
from datetime import datetime
from typing import List

from rmlab._data.enums import ScenarioDayStatus, ScenarioState
from rmlab._data.types import DateFormat, DateTimeSecFormat


@dataclass
class Scenario:
    """Base class for dataclasses holding per-scenario information.

    Args:
        id (int): Scenario identifier

    """

    id: int

    def __post_init__(self):
        self.id = int(self.id)

    def __repr__(self) -> str:
        return self.__dict__


@dataclass
class ScenarioDates(Scenario):
    """Set of dates and status characteristic to a scenario.

    Args:
        day_status (ScenarioDayStatus): Intra day scenario status
        current (datetime): Current date upto days (year-month-day)
        next (datetime): Date at which simulation would stop, upto days (year-month-day)
        first_flight_load (datetime): Date at which the first flight is put on sale, upto days (year-month-day)
        last_flight_departure (datetime): Date at which the last flight departs, upto days (year-month-day)
        checkpoints (datetime): List of dates at which simulation would stop, upto days (year-month-day)
        state (ScenarioState): Global state of scenario
    """

    day_status: ScenarioDayStatus
    current: datetime
    next: datetime
    first_flight_load: datetime
    last_flight_departure: datetime
    checkpoints: List[datetime]
    upload_time: datetime
    state: ScenarioState

    def __post_init__(self):

        self.day_status = ScenarioDayStatus.str_to_enum_value(self.day_status)
        self.current = datetime.strptime(self.current, DateFormat)
        self.next = datetime.strptime(self.next, DateFormat)
        self.first_flight_load = datetime.strptime(self.first_flight_load, DateFormat)
        self.last_flight_departure = datetime.strptime(
            self.last_flight_departure, DateFormat
        )
        self.checkpoints = [
            datetime.strptime(chp, DateFormat) for chp in self.checkpoints
        ]
        self.upload_time = datetime.strptime(self.upload_time, DateTimeSecFormat)


@dataclass
class ItemsCount(Scenario):
    """Counters of items of bounded categories.

    Args:
        aircraft (int): Number of referenced aircrafts in scenario
        airline (int): Number of referenced airlines in scenario
        airport (int): Number of referenced airports in scenario
        city (int): Number of referenced cities in scenario
        country (int): Number of referenced countries in scenario
        cityroute (int): Number of cityroutes in scenario
        citysector (int): Number of citysectors in scenario
        route (int): Number of routes in scenario
        sector (int): Number of sectors in scenario
        pmodel (int): Number of parametric models in scenario
    """

    aircraft: int
    airline: int
    airport: int
    city: int
    cityroute: int
    citysector: int
    country: int
    pmodel: int
    route: int
    sector: int


@dataclass
class SchedulesCount(Scenario):
    """Counters of flights schedules in a given scenario.

    Args:
        past (int): Number of schedules with all flights departed at current date
        live (int): Number of schedules with flights on sale at current date
        pending (int): Number of schedules with all flights not yet on sale at current date
        total (int): Total number of schedules
    """

    past: int
    live: int
    pending: int
    total: int = field(init=False)

    def __post_init__(self):
        self.total = self.past + self.live + self.pending


@dataclass
class FlightsCount(Scenario):
    """Counters of flights in a given scenario.

    Args:
        past (int): Number of flights departed at current date
        live (int): Number of flights on sale at current date
        pending (int): Number of flights not departed and not yet on sale at current date
        total (int): Total number of flights (schedules)
    """

    past: int
    live: int
    pending: int
    total: int = field(init=False)

    def __post_init__(self):
        self.total = self.past + self.live + self.pending
